# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

import mock
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils
from taskflow.types import failure

from octavia.common import constants
from octavia.common import data_models as o_data_models
from octavia.controller.worker.tasks import network_tasks
from octavia.network import base as net_base
from octavia.network import data_models
from octavia.tests.common import constants as t_constants
import octavia.tests.unit.base as base


AMPHORA_ID = 7
COMPUTE_ID = uuidutils.generate_uuid()
PORT_ID = uuidutils.generate_uuid()
SUBNET_ID = uuidutils.generate_uuid()
NETWORK_ID = uuidutils.generate_uuid()
IP_ADDRESS = "172.24.41.1"
VIP = o_data_models.Vip(port_id=t_constants.MOCK_PORT_ID,
                        subnet_id=t_constants.MOCK_SUBNET_ID,
                        qos_policy_id=t_constants.MOCK_QOS_POLICY_ID1)
VIP2 = o_data_models.Vip(port_id=t_constants.MOCK_PORT_ID2,
                         subnet_id=t_constants.MOCK_SUBNET_ID2,
                         qos_policy_id=t_constants.MOCK_QOS_POLICY_ID2)
LB = o_data_models.LoadBalancer(vip=VIP)
LB2 = o_data_models.LoadBalancer(vip=VIP2)
FIRST_IP = {"ip_address": IP_ADDRESS, "subnet_id": SUBNET_ID}
FIXED_IPS = [FIRST_IP]
INTERFACE = data_models.Interface(id=uuidutils.generate_uuid(),
                                  compute_id=COMPUTE_ID, fixed_ips=FIXED_IPS,
                                  port_id=PORT_ID)
AMPS_DATA = [o_data_models.Amphora(id=t_constants.MOCK_AMP_ID1,
                                   vrrp_port_id=t_constants.MOCK_VRRP_PORT_ID1,
                                   vrrp_ip=t_constants.MOCK_VRRP_IP1),
             o_data_models.Amphora(id=t_constants.MOCK_AMP_ID2,
                                   vrrp_port_id=t_constants.MOCK_VRRP_PORT_ID2,
                                   vrrp_ip=t_constants.MOCK_VRRP_IP2)
             ]
UPDATE_DICT = {constants.LOADBALANCER_TOPOLOGY: None}


class TestException(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


@mock.patch('octavia.common.utils.get_network_driver')
class TestNetworkTasks(base.TestCase):
    def setUp(self):
        network_tasks.LOG = mock.MagicMock()
        self.amphora_mock = mock.MagicMock()
        self.load_balancer_mock = mock.MagicMock()
        self.vip_mock = mock.MagicMock()
        self.vip_mock.subnet_id = SUBNET_ID
        self.load_balancer_mock.vip = self.vip_mock
        self.load_balancer_mock.amphorae = []
        self.amphora_mock.id = AMPHORA_ID
        self.amphora_mock.compute_id = COMPUTE_ID
        self.amphora_mock.status = constants.AMPHORA_ALLOCATED
        conf = oslo_fixture.Config(cfg.CONF)
        conf.config(group="controller_worker", amp_boot_network_list=['netid'])

        super(TestNetworkTasks, self).setUp()

    def test_calculate_delta(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        EMPTY = {}
        empty_deltas = {self.amphora_mock.id: data_models.Delta(
            amphora_id=self.amphora_mock.id,
            compute_id=self.amphora_mock.compute_id,
            add_nics=[],
            delete_nics=[])}

        def _interface(network_id):
            return [data_models.Interface(network_id=network_id)]

        net = network_tasks.CalculateDelta()

        self.assertEqual(EMPTY, net.execute(self.load_balancer_mock))

        self.amphora_mock.load_balancer = self.load_balancer_mock
        self.load_balancer_mock.amphorae = [self.amphora_mock]
        self.load_balancer_mock.pools = []

        self.assertEqual(empty_deltas, net.execute(self.load_balancer_mock))
        mock_driver.get_plugged_networks.assert_called_once_with(COMPUTE_ID)

        pool_mock = mock.MagicMock()
        self.load_balancer_mock.pools = [pool_mock]
        pool_mock.members = []
        self.assertEqual(empty_deltas, net.execute(self.load_balancer_mock))

        member_mock = mock.MagicMock()
        pool_mock.members = [member_mock]
        member_mock.subnet_id = 1

        mock_driver.get_subnet.reset_mock()
        mock_driver.get_subnet.return_value = data_models.Subnet(id=2,
                                                                 network_id=3)

        ndm = data_models.Delta(amphora_id=self.amphora_mock.id,
                                compute_id=self.amphora_mock.compute_id,
                                add_nics=_interface(2),
                                delete_nics=[])
        self.assertEqual({self.amphora_mock.id: ndm},
                         net.execute(self.load_balancer_mock))

        vrrp_port_call = mock.call(self.amphora_mock.vrrp_port_id)
        mock_driver.get_port.assert_has_calls([vrrp_port_call])
        # For some reason we call calculate_delta three times?
        self.assertEqual(3, mock_driver.get_port.call_count)

        member_subnet_call = mock.call(member_mock.subnet_id)
        mock_driver.get_subnet.assert_has_calls([member_subnet_call])
        self.assertEqual(1, mock_driver.get_subnet.call_count)

        mock_driver.get_plugged_networks.return_value = _interface(2)
        self.assertEqual(empty_deltas, net.execute(self.load_balancer_mock))

        mock_driver.get_plugged_networks.return_value = _interface(3)
        ndm = data_models.Delta(amphora_id=self.amphora_mock.id,
                                compute_id=self.amphora_mock.compute_id,
                                add_nics=_interface(2),
                                delete_nics=_interface(3))
        self.assertEqual({self.amphora_mock.id: ndm},
                         net.execute(self.load_balancer_mock))

        pool_mock.members = []
        mock_driver.get_plugged_networks.return_value = _interface(2)
        ndm = data_models.Delta(amphora_id=self.amphora_mock.id,
                                compute_id=self.amphora_mock.compute_id,
                                add_nics=[],
                                delete_nics=_interface(2))
        self.assertEqual({self.amphora_mock.id: ndm},
                         net.execute(self.load_balancer_mock))

    def test_get_plumbed_networks(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        mock_driver.get_plugged_networks.side_effect = [['blah']]
        net = network_tasks.GetPlumbedNetworks()

        self.assertEqual(['blah'], net.execute(self.amphora_mock))
        mock_driver.get_plugged_networks.assert_called_once_with(
            COMPUTE_ID)

    def test_plug_networks(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        def _interface(network_id):
            return [data_models.Interface(network_id=network_id)]

        net = network_tasks.PlugNetworks()

        net.execute(self.amphora_mock, None)
        self.assertFalse(mock_driver.plug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=[])
        net.execute(self.amphora_mock, delta)
        self.assertFalse(mock_driver.plug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=_interface(1),
                                  delete_nics=[])
        net.execute(self.amphora_mock, delta)
        mock_driver.plug_network.assert_called_once_with(COMPUTE_ID, 1)

        # revert
        net.revert(self.amphora_mock, None)
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=[])
        net.revert(self.amphora_mock, delta)
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=_interface(1),
                                  delete_nics=[])
        net.revert(self.amphora_mock, delta)
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = net_base.NetworkNotFound
        net.revert(self.amphora_mock, delta)  # No exception
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = TestException('test')
        self.assertRaises(TestException,
                          net.revert,
                          self.amphora_mock,
                          delta)
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

    def test_unplug_networks(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        def _interface(network_id):
            return [data_models.Interface(network_id=network_id)]

        net = network_tasks.UnPlugNetworks()

        net.execute(self.amphora_mock, None)
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=[])
        net.execute(self.amphora_mock, delta)
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=_interface(1))
        net.execute(self.amphora_mock, delta)
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = net_base.NetworkNotFound
        net.execute(self.amphora_mock, delta)  # No exception
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        # Do a test with a general exception in case behavior changes
        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = Exception()
        net.execute(self.amphora_mock, delta)  # No exception
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

    def test_get_member_ports(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        def _interface(port_id):
            return [data_models.Interface(port_id=port_id)]

        net_task = network_tasks.GetMemberPorts()
        net_task.execute(LB, self.amphora_mock)
        mock_driver.get_port.assert_called_once_with(t_constants.MOCK_PORT_ID)
        mock_driver.get_plugged_networks.assert_called_once_with(COMPUTE_ID)

        mock_driver.reset_mock()
        net_task = network_tasks.GetMemberPorts()
        mock_driver.get_plugged_networks.return_value = _interface(1)
        mock_driver.get_port.side_effect = [
            data_models.Port(network_id=NETWORK_ID),
            data_models.Port(network_id=NETWORK_ID)]
        net_task.execute(self.load_balancer_mock, self.amphora_mock)
        self.assertEqual(2, mock_driver.get_port.call_count)
        self.assertFalse(mock_driver.get_network.called)

        mock_driver.reset_mock()
        port_mock = mock.MagicMock()
        fixed_ip_mock = mock.MagicMock()
        fixed_ip_mock.subnet_id = 1
        port_mock.fixed_ips = [fixed_ip_mock]
        net_task = network_tasks.GetMemberPorts()
        mock_driver.get_plugged_networks.return_value = _interface(1)
        mock_driver.get_port.side_effect = [
            data_models.Port(network_id=NETWORK_ID), port_mock]
        ports = net_task.execute(self.load_balancer_mock, self.amphora_mock)
        mock_driver.get_subnet.assert_called_once_with(1)
        self.assertEqual([port_mock], ports)

    def test_handle_network_delta(self, mock_get_net_driver):
        mock_net_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_net_driver

        nic1 = mock.MagicMock()
        nic1.network_id = uuidutils.generate_uuid()
        nic2 = mock.MagicMock()
        nic2.network_id = uuidutils.generate_uuid()
        interface1 = mock.MagicMock()
        interface1.port_id = uuidutils.generate_uuid()
        port1 = mock.MagicMock()
        port1.network_id = uuidutils.generate_uuid()
        fixed_ip = mock.MagicMock()
        fixed_ip.subnet_id = uuidutils.generate_uuid()
        port1.fixed_ips = [fixed_ip]
        subnet = mock.MagicMock()
        network = mock.MagicMock()

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[nic1],
                                  delete_nics=[nic2, nic2, nic2])

        mock_net_driver.plug_network.return_value = interface1
        mock_net_driver.get_port.return_value = port1
        mock_net_driver.get_network.return_value = network
        mock_net_driver.get_subnet.return_value = subnet

        mock_net_driver.unplug_network.side_effect = [
            None, net_base.NetworkNotFound, Exception]

        handle_net_delta_obj = network_tasks.HandleNetworkDelta()
        result = handle_net_delta_obj.execute(self.amphora_mock, delta)

        mock_net_driver.plug_network.assert_called_once_with(
            self.amphora_mock.compute_id, nic1.network_id)
        mock_net_driver.get_port.assert_called_once_with(interface1.port_id)
        mock_net_driver.get_network.assert_called_once_with(port1.network_id)
        mock_net_driver.get_subnet.assert_called_once_with(fixed_ip.subnet_id)

        self.assertEqual({self.amphora_mock.id: [port1]}, result)

        mock_net_driver.unplug_network.assert_called_with(
            self.amphora_mock.compute_id, nic2.network_id)

        # Revert
        delta2 = data_models.Delta(amphora_id=self.amphora_mock.id,
                                   compute_id=self.amphora_mock.compute_id,
                                   add_nics=[nic1, nic1],
                                   delete_nics=[nic2, nic2, nic2])

        mock_net_driver.unplug_network.reset_mock()
        handle_net_delta_obj.revert(
            failure.Failure.from_exception(Exception('boom')), None, None)
        mock_net_driver.unplug_network.assert_not_called()

        mock_net_driver.unplug_network.reset_mock()
        handle_net_delta_obj.revert(None, None, None)
        mock_net_driver.unplug_network.assert_not_called()

        mock_net_driver.unplug_network.reset_mock()
        handle_net_delta_obj.revert(None, None, delta2)

    def test_handle_network_deltas(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        def _interface(network_id):
            return [data_models.Interface(network_id=network_id)]

        net = network_tasks.HandleNetworkDeltas()

        net.execute({})
        self.assertFalse(mock_driver.plug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=[])
        net.execute({self.amphora_mock.id: delta})
        self.assertFalse(mock_driver.plug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=_interface(1),
                                  delete_nics=[])
        net.execute({self.amphora_mock.id: delta})
        mock_driver.plug_network.assert_called_once_with(COMPUTE_ID, 1)

        # revert
        net.execute({self.amphora_mock.id: delta})
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=[])
        net.execute({self.amphora_mock.id: delta})
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=_interface(1),
                                  delete_nics=[])

        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = net_base.NetworkNotFound

        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = TestException('test')
        self.assertRaises(TestException, net.revert, mock.ANY,
                          {self.amphora_mock.id: delta})
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        mock_driver.reset_mock()
        net.execute({})
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=[])
        net.execute({self.amphora_mock.id: delta})
        self.assertFalse(mock_driver.unplug_network.called)

        delta = data_models.Delta(amphora_id=self.amphora_mock.id,
                                  compute_id=self.amphora_mock.compute_id,
                                  add_nics=[],
                                  delete_nics=_interface(1))
        net.execute({self.amphora_mock.id: delta})
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = net_base.NetworkNotFound
        net.execute({self.amphora_mock.id: delta})
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

        # Do a test with a general exception in case behavior changes
        mock_driver.reset_mock()
        mock_driver.unplug_network.side_effect = Exception()
        net.execute({self.amphora_mock.id: delta})
        mock_driver.unplug_network.assert_called_once_with(COMPUTE_ID, 1)

    def test_plug_vip(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        net = network_tasks.PlugVIP()

        mock_driver.plug_vip.return_value = ["vip"]

        data = net.execute(LB)
        mock_driver.plug_vip.assert_called_once_with(LB, LB.vip)
        self.assertEqual(["vip"], data)

        # revert
        net.revert(["vip"], LB)
        mock_driver.unplug_vip.assert_called_once_with(LB, LB.vip)

        # revert with exception
        mock_driver.reset_mock()
        mock_driver.unplug_vip.side_effect = Exception('UnplugVipException')
        net.revert(["vip"], LB)
        mock_driver.unplug_vip.assert_called_once_with(LB, LB.vip)

    @mock.patch('octavia.controller.worker.task_utils.TaskUtils.'
                'get_current_loadbalancer_from_db')
    def test_apply_qos_on_creation(self, mock_get_lb_db, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        net = network_tasks.ApplyQos()
        mock_get_lb_db.return_value = LB

        # execute
        UPDATE_DICT[
            constants.LOADBALANCER_TOPOLOGY] = constants.TOPOLOGY_SINGLE
        update_dict = UPDATE_DICT
        net.execute(LB, [AMPS_DATA[0]], update_dict)
        mock_driver.apply_qos_on_port.assert_called_once_with(
            VIP.qos_policy_id, AMPS_DATA[0].vrrp_port_id)
        self.assertEqual(1, mock_driver.apply_qos_on_port.call_count)
        standby_topology = constants.TOPOLOGY_ACTIVE_STANDBY
        mock_driver.reset_mock()
        update_dict[
            constants.LOADBALANCER_TOPOLOGY] = standby_topology
        net.execute(LB, AMPS_DATA, update_dict)
        mock_driver.apply_qos_on_port.assert_called_with(
            t_constants.MOCK_QOS_POLICY_ID1, mock.ANY)
        self.assertEqual(2, mock_driver.apply_qos_on_port.call_count)

        # revert
        mock_driver.reset_mock()
        update_dict = UPDATE_DICT
        net.revert(None, LB, [AMPS_DATA[0]], update_dict)
        self.assertEqual(0, mock_driver.apply_qos_on_port.call_count)
        mock_driver.reset_mock()
        update_dict[
            constants.LOADBALANCER_TOPOLOGY] = standby_topology
        net.revert(None, LB, AMPS_DATA, update_dict)
        self.assertEqual(0, mock_driver.apply_qos_on_port.call_count)

    @mock.patch('octavia.controller.worker.task_utils.TaskUtils.'
                'get_current_loadbalancer_from_db')
    def test_apply_qos_on_update(self, mock_get_lb_db, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        net = network_tasks.ApplyQos()
        null_qos_vip = o_data_models.Vip(qos_policy_id=None)
        null_qos_lb = o_data_models.LoadBalancer(
            vip=null_qos_vip, topology=constants.TOPOLOGY_SINGLE,
            amphorae=[AMPS_DATA[0]])

        tmp_vip_object = o_data_models.Vip(
            qos_policy_id=t_constants.MOCK_QOS_POLICY_ID1)
        tmp_lb = o_data_models.LoadBalancer(
            vip=tmp_vip_object, topology=constants.TOPOLOGY_SINGLE,
            amphorae=[AMPS_DATA[0]])

        # execute
        update_dict = {'description': 'fool'}
        net.execute(tmp_lb, update_dict=update_dict)
        mock_driver.apply_qos_on_port.assert_called_once_with(
            t_constants.MOCK_QOS_POLICY_ID1, AMPS_DATA[0].vrrp_port_id)
        self.assertEqual(1, mock_driver.apply_qos_on_port.call_count)

        mock_driver.reset_mock()
        update_dict = {'vip': {'qos_policy_id': None}}
        net.execute(null_qos_lb, update_dict=update_dict)
        mock_driver.apply_qos_on_port.assert_called_once_with(
            None, AMPS_DATA[0].vrrp_port_id)
        self.assertEqual(1, mock_driver.apply_qos_on_port.call_count)

        mock_driver.reset_mock()
        update_dict = {'name': '123'}
        net.execute(null_qos_lb, update_dict=update_dict)
        self.assertEqual(0, mock_driver.apply_qos_on_port.call_count)

        mock_driver.reset_mock()
        update_dict = {'description': 'fool'}
        tmp_lb.amphorae = AMPS_DATA
        tmp_lb.topology = constants.TOPOLOGY_ACTIVE_STANDBY
        net.execute(tmp_lb, update_dict=update_dict)
        mock_driver.apply_qos_on_port.assert_called_with(
            t_constants.MOCK_QOS_POLICY_ID1, mock.ANY)
        self.assertEqual(2, mock_driver.apply_qos_on_port.call_count)

        # revert
        mock_driver.reset_mock()
        tmp_lb.amphorae = [AMPS_DATA[0]]
        tmp_lb.topology = constants.TOPOLOGY_SINGLE
        update_dict = {'description': 'fool'}
        mock_get_lb_db.return_value = tmp_lb
        net.revert(None, tmp_lb, update_dict=update_dict)
        self.assertEqual(0, mock_driver.apply_qos_on_port.call_count)

        mock_driver.reset_mock()
        update_dict = {'vip': {'qos_policy_id': None}}
        ori_lb_db = LB2
        ori_lb_db.amphorae = [AMPS_DATA[0]]
        mock_get_lb_db.return_value = ori_lb_db
        net.revert(None, null_qos_lb, update_dict=update_dict)
        mock_driver.apply_qos_on_port.assert_called_once_with(
            t_constants.MOCK_QOS_POLICY_ID2, AMPS_DATA[0].vrrp_port_id)
        self.assertEqual(1, mock_driver.apply_qos_on_port.call_count)

        mock_driver.reset_mock()
        update_dict = {'vip': {
            'qos_policy_id': t_constants.MOCK_QOS_POLICY_ID2}}
        tmp_lb.amphorae = AMPS_DATA
        tmp_lb.topology = constants.TOPOLOGY_ACTIVE_STANDBY
        ori_lb_db = LB2
        ori_lb_db.amphorae = [AMPS_DATA[0]]
        mock_get_lb_db.return_value = ori_lb_db
        net.revert(None, tmp_lb, update_dict=update_dict)
        mock_driver.apply_qos_on_port.assert_called_with(
            t_constants.MOCK_QOS_POLICY_ID2, mock.ANY)
        self.assertEqual(2, mock_driver.apply_qos_on_port.call_count)

    def test_unplug_vip(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        net = network_tasks.UnplugVIP()

        net.execute(LB)
        mock_driver.unplug_vip.assert_called_once_with(LB, LB.vip)

    def test_allocate_vip(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        net = network_tasks.AllocateVIP()

        mock_driver.allocate_vip.return_value = LB.vip

        mock_driver.reset_mock()
        self.assertEqual(LB.vip, net.execute(LB))
        mock_driver.allocate_vip.assert_called_once_with(LB)

        # revert
        vip_mock = mock.MagicMock()
        net.revert(vip_mock, LB)
        mock_driver.deallocate_vip.assert_called_once_with(vip_mock)

        # revert exception
        mock_driver.reset_mock()
        mock_driver.deallocate_vip.side_effect = Exception('DeallVipException')
        vip_mock = mock.MagicMock()
        net.revert(vip_mock, LB)
        mock_driver.deallocate_vip.assert_called_once_with(vip_mock)

    def test_deallocate_vip(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        net = network_tasks.DeallocateVIP()
        vip = o_data_models.Vip()
        lb = o_data_models.LoadBalancer(vip=vip)
        net.execute(lb)
        mock_driver.deallocate_vip.assert_called_once_with(lb.vip)

    def test_update_vip(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        vip = o_data_models.Vip()
        lb = o_data_models.LoadBalancer(vip=vip)
        net_task = network_tasks.UpdateVIP()
        net_task.execute(lb)
        mock_driver.update_vip.assert_called_once_with(lb)

    def test_update_vip_for_delete(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        vip = o_data_models.Vip()
        lb = o_data_models.LoadBalancer(vip=vip)
        net_task = network_tasks.UpdateVIPForDelete()
        net_task.execute(lb)
        mock_driver.update_vip.assert_called_once_with(lb, for_delete=True)

    def test_get_amphorae_network_configs(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        lb = o_data_models.LoadBalancer()
        net_task = network_tasks.GetAmphoraeNetworkConfigs()
        net_task.execute(lb)
        mock_driver.get_network_configs.assert_called_once_with(lb)

    def test_failover_preparation_for_amphora(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        failover = network_tasks.FailoverPreparationForAmphora()
        amphora = o_data_models.Amphora(id=AMPHORA_ID,
                                        lb_network_ip=IP_ADDRESS)
        failover.execute(amphora)
        mock_driver.failover_preparation.assert_called_once_with(amphora)

    def test_retrieve_portids_on_amphora_except_lb_network(
            self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        def _interface(port_id):
            return [data_models.Interface(port_id=port_id)]

        net_task = network_tasks.RetrievePortIDsOnAmphoraExceptLBNetwork()
        amphora = o_data_models.Amphora(id=AMPHORA_ID, compute_id=COMPUTE_ID,
                                        lb_network_ip=IP_ADDRESS)

        mock_driver.get_plugged_networks.return_value = []
        net_task.execute(amphora)
        mock_driver.get_plugged_networks.assert_called_once_with(
            compute_id=COMPUTE_ID)
        self.assertFalse(mock_driver.get_port.called)

        mock_driver.reset_mock()
        net_task = network_tasks.RetrievePortIDsOnAmphoraExceptLBNetwork()
        mock_driver.get_plugged_networks.return_value = _interface(1)
        net_task.execute(amphora)
        mock_driver.get_port.assert_called_once_with(port_id=1)

        mock_driver.reset_mock()
        net_task = network_tasks.RetrievePortIDsOnAmphoraExceptLBNetwork()
        port_mock = mock.MagicMock()
        fixed_ip_mock = mock.MagicMock()
        fixed_ip_mock.ip_address = IP_ADDRESS
        port_mock.fixed_ips = [fixed_ip_mock]
        mock_driver.get_plugged_networks.return_value = _interface(1)
        mock_driver.get_port.return_value = port_mock
        ports = net_task.execute(amphora)
        self.assertEqual([], ports)

        mock_driver.reset_mock()
        net_task = network_tasks.RetrievePortIDsOnAmphoraExceptLBNetwork()
        port_mock = mock.MagicMock()
        fixed_ip_mock = mock.MagicMock()
        fixed_ip_mock.ip_address = "172.17.17.17"
        port_mock.fixed_ips = [fixed_ip_mock]
        mock_driver.get_plugged_networks.return_value = _interface(1)
        mock_driver.get_port.return_value = port_mock
        ports = net_task.execute(amphora)
        self.assertEqual(1, len(ports))

    def test_plug_ports(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        amphora = mock.MagicMock()
        port1 = mock.MagicMock()
        port2 = mock.MagicMock()

        plugports = network_tasks.PlugPorts()
        plugports.execute(amphora, [port1, port2])

        mock_driver.plug_port.assert_any_call(amphora, port1)
        mock_driver.plug_port.assert_any_call(amphora, port2)

    def test_plug_vip_port(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver
        vrrp_port = mock.MagicMock()

        amphorae_network_config = mock.MagicMock()
        amphorae_network_config.get().vrrp_port = vrrp_port

        plugvipport = network_tasks.PlugVIPPort()
        plugvipport.execute(self.amphora_mock, amphorae_network_config)
        mock_driver.plug_port.assert_any_call(self.amphora_mock, vrrp_port)

        # test revert
        plugvipport.revert(None, self.amphora_mock, amphorae_network_config)
        mock_driver.unplug_port.assert_any_call(self.amphora_mock, vrrp_port)

    def test_wait_for_port_detach(self, mock_get_net_driver):
        mock_driver = mock.MagicMock()
        mock_get_net_driver.return_value = mock_driver

        amphora = o_data_models.Amphora(id=AMPHORA_ID,
                                        lb_network_ip=IP_ADDRESS)

        waitforportdetach = network_tasks.WaitForPortDetach()
        waitforportdetach.execute(amphora)

        mock_driver.wait_for_port_detach.assert_called_once_with(amphora)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from octavia.compute.drivers.noop_driver import driver
import octavia.tests.unit.base as base


class TestNoopComputeDriver(base.TestCase):
    FAKE_UUID_1 = uuidutils.generate_uuid()
    FAKE_UUID_2 = uuidutils.generate_uuid()
    FAKE_UUID_3 = uuidutils.generate_uuid()
    FAKE_UUID_4 = uuidutils.generate_uuid()
    FAKE_UUID_5 = uuidutils.generate_uuid()
    FAKE_UUID_6 = uuidutils.generate_uuid()

    def setUp(self):
        super(TestNoopComputeDriver, self).setUp()
        self.driver = driver.NoopComputeDriver()

        self.name = "amphora_name"
        self.amphora_flavor = "m1.tiny"
        self.image_id = self.FAKE_UUID_1
        self.image_tag = "faketag"
        self.image_owner = self.FAKE_UUID_2
        self.key_name = "key_name"
        self.sec_groups = ["default"]
        self.network_ids = [self.FAKE_UUID_3]
        self.confdrivefiles = ["config_driver_file1"]
        self.user_data = "user_data"
        self.amphora_id = self.FAKE_UUID_4
        self.loadbalancer_id = self.FAKE_UUID_5
        self.server_group_policy = 'anti-affinity'
        self.server_group_name = 'my_server_group'
        self.server_group_id = self.FAKE_UUID_6
        self.port_ids = ['port-id-1']

    def test_build(self):
        self.driver.build(self.name, self.amphora_flavor,
                          self.image_id, self.image_tag, self.image_owner,
                          self.key_name, self.sec_groups, self.network_ids,
                          self.confdrivefiles, self.user_data, self.port_ids,
                          self.server_group_id)

        self.assertEqual((self.name, self.amphora_flavor,
                          self.image_id, self.image_tag, self.image_owner,
                          self.key_name, self.sec_groups, self.network_ids,
                          self.confdrivefiles, self.user_data, self.port_ids,
                          self.server_group_id, 'build'),
                         self.driver.driver.computeconfig[(
                             self.name,
                             self.amphora_flavor,
                             self.image_id,
                             self.image_tag,
                             self.image_owner,
                             self.key_name,
                             self.user_data,
                             self.server_group_id
                         )])

    def test_delete(self):
        self.driver.delete(self.amphora_id)
        self.assertEqual((self.amphora_id, 'delete'),
                         self.driver.driver.computeconfig[
                             self.amphora_id])

    def test_status(self):
        self.driver.status(self.amphora_id)
        self.assertEqual((self.amphora_id, 'status'),
                         self.driver.driver.computeconfig[
                             self.amphora_id])

    def test_get_amphora(self):
        self.driver.get_amphora(self.amphora_id)
        self.assertEqual((self.amphora_id, 'get_amphora'),
                         self.driver.driver.computeconfig[
                             self.amphora_id])

    def test_create_server_group(self):
        self.driver.create_server_group(self.server_group_name,
                                        self.server_group_policy)
        self.assertEqual((self.server_group_name, self.server_group_policy,
                          'create'),
                         self.driver.driver.computeconfig[
                             self.server_group_name, self.server_group_policy])

    def test_delete_server_group(self):
        self.driver.delete_server_group(self.server_group_id)
        self.assertEqual((self.server_group_id, 'delete'),
                         self.driver.driver.computeconfig[
                             self.server_group_id])



#    Copyright 2014 Rackspace
#    Copyright 2016 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
This is just a handler that will simulate successful operations a controller
should perform.  There is nothing useful about this other than database
entity status management.
"""

import threading
import time

from oslo_log import log as logging

from octavia.api.handlers import abstract_handler
from octavia.common import constants
from octavia.common import data_models
from octavia.db import api as db_api
import octavia.db.repositories as repos

LOG = logging.getLogger(__name__)
ASYNC_TIME = 1


def validate_input(expected, actual):
    if not isinstance(actual, expected):
        raise InvalidHandlerInputObject(obj_type=actual.__class__)


def simulate_controller(data_model, delete=False, update=False, create=False,
                        batch_update=False):
    """Simulates a successful controller operator for a data model.

    :param data_model: data model to simulate controller operation
    :param delete: deletes from the database
    """
    repo = repos.Repositories()

    def member_controller(member, delete=False, update=False, create=False,
                          batch_update=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for member...")

        db_mem = None
        if delete:
            db_mem = repo.member.get(db_api.get_session(), id=member.id)
            repo.member.delete(db_api.get_session(), id=member.id)
        elif update:
            db_mem = repo.member.get(db_api.get_session(), id=member.id)
            member_dict = member.to_dict()
            member_dict['operating_status'] = db_mem.operating_status
            repo.member.update(db_api.get_session(), member.id, **member_dict)
        elif create:
            repo.member.update(db_api.get_session(), member.id,
                               operating_status=constants.ONLINE)
        elif batch_update:
            members = member
            for member in members:
                repo.member.update(db_api.get_session(), member.id,
                                   operating_status=constants.ONLINE)
        listeners = []
        if db_mem:
            for listener in db_mem.pool.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        if member.pool.listeners:
            for listener in member.pool.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        if listeners:
            for listener in listeners:
                repo.listener.update(db_api.get_session(), listener.id,
                                     operating_status=constants.ONLINE,
                                     provisioning_status=constants.ACTIVE)
        repo.load_balancer.update(db_api.get_session(),
                                  member.pool.load_balancer.id,
                                  operating_status=constants.ONLINE,
                                  provisioning_status=constants.ACTIVE)
        LOG.info("Simulated Controller Handler Thread Complete")

    def l7policy_controller(l7policy, delete=False, update=False,
                            create=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for l7policy...")

        db_l7policy = None
        if delete:
            db_l7policy = repo.l7policy.get(db_api.get_session(),
                                            id=l7policy.id)
            repo.l7policy.delete(db_api.get_session(), id=l7policy.id)
        elif update:
            db_l7policy = repo.l7policy.get(db_api.get_session(),
                                            id=l7policy.id)
            l7policy_dict = l7policy.to_dict()
            repo.l7policy.update(db_api.get_session(), l7policy.id,
                                 **l7policy_dict)
        elif create:
            db_l7policy = repo.l7policy.create(db_api.get_session(),
                                               **l7policy.to_dict())
        if db_l7policy.listener:
            repo.listener.update(db_api.get_session(), db_l7policy.listener.id,
                                 operating_status=constants.ONLINE,
                                 provisioning_status=constants.ACTIVE)
            repo.load_balancer.update(db_api.get_session(),
                                      db_l7policy.listener.load_balancer.id,
                                      operating_status=constants.ONLINE,
                                      provisioning_status=constants.ACTIVE)
        LOG.info("Simulated Controller Handler Thread Complete")

    def l7rule_controller(l7rule, delete=False, update=False, create=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for l7rule...")

        db_l7rule = None
        if delete:
            db_l7rule = repo.l7rule.get(db_api.get_session(), id=l7rule.id)
            repo.l7rule.delete(db_api.get_session(), id=l7rule.id)
        elif update:
            db_l7rule = repo.l7rule.get(db_api.get_session(), id=l7rule.id)
            l7rule_dict = l7rule.to_dict()
            repo.l7rule.update(db_api.get_session(), l7rule.id, **l7rule_dict)
        elif create:
            l7rule_dict = l7rule.to_dict()
            db_l7rule = repo.l7rule.create(db_api.get_session(), **l7rule_dict)
        if db_l7rule.l7policy.listener:
            listener = db_l7rule.l7policy.listener
            repo.listener.update(db_api.get_session(), listener.id,
                                 operating_status=constants.ONLINE,
                                 provisioning_status=constants.ACTIVE)
            repo.load_balancer.update(db_api.get_session(),
                                      listener.load_balancer.id,
                                      operating_status=constants.ONLINE,
                                      provisioning_status=constants.ACTIVE)
        LOG.info("Simulated Controller Handler Thread Complete")

    def health_monitor_controller(health_monitor, delete=False, update=False,
                                  create=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for health monitor...")

        db_hm = None
        if delete:
            db_hm = repo.health_monitor.get(db_api.get_session(),
                                            pool_id=health_monitor.pool.id)
            repo.health_monitor.delete(db_api.get_session(),
                                       pool_id=health_monitor.pool.id)
        elif update:
            db_hm = repo.health_monitor.get(db_api.get_session(),
                                            pool_id=health_monitor.pool_id)
            hm_dict = health_monitor.to_dict()
            hm_dict['operating_status'] = db_hm.operating_status()
            repo.health_monitor.update(db_api.get_session(), **hm_dict)
        elif create:
            repo.pool.update(db_api.get_session(), health_monitor.pool_id,
                             operating_status=constants.ONLINE)
        listeners = []
        if db_hm:
            for listener in db_hm.pool.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        if health_monitor.pool.listeners:
            for listener in health_monitor.pool.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        if listeners:
            for listener in listeners:
                repo.test_and_set_lb_and_listener_prov_status(
                    db_api.get_session(),
                    health_monitor.pool.load_balancer.id,
                    listener.id, constants.ACTIVE,
                    constants.ACTIVE)
                repo.listener.update(db_api.get_session(),
                                     listener.id,
                                     operating_status=constants.ONLINE,
                                     provisioning_status=constants.ACTIVE)
        repo.load_balancer.update(
            db_api.get_session(),
            health_monitor.pool.load_balancer.id,
            operating_status=constants.ONLINE,
            provisioning_status=constants.ACTIVE)
        LOG.info("Simulated Controller Handler Thread Complete")

    def pool_controller(pool, delete=False, update=False, create=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for pool...")

        db_pool = None
        if delete:
            db_pool = repo.pool.get(db_api.get_session(), id=pool.id)
            repo.pool.delete(db_api.get_session(), id=pool.id)
        elif update:
            db_pool = repo.pool.get(db_api.get_session(), id=pool.id)
            pool_dict = pool.to_dict()
            pool_dict['operating_status'] = db_pool.operating_status
            repo.update_pool_and_sp(db_api.get_session(), pool.id, pool_dict)
        elif create:
            repo.pool.update(db_api.get_session(), pool.id,
                             operating_status=constants.ONLINE)
        listeners = []
        if db_pool:
            for listener in db_pool.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        if pool.listeners:
            for listener in pool.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        if listeners:
            for listener in listeners:
                repo.listener.update(db_api.get_session(), listener.id,
                                     operating_status=constants.ONLINE,
                                     provisioning_status=constants.ACTIVE)
        repo.load_balancer.update(db_api.get_session(),
                                  pool.load_balancer.id,
                                  operating_status=constants.ONLINE,
                                  provisioning_status=constants.ACTIVE)
        LOG.info("Simulated Controller Handler Thread Complete")

    def listener_controller(listener, delete=False, update=False,
                            create=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for listener...")

        if delete:
            repo.listener.update(db_api.get_session(), listener.id,
                                 operating_status=constants.OFFLINE,
                                 provisioning_status=constants.DELETED)
        elif update:
            db_listener = repo.listener.get(db_api.get_session(),
                                            id=listener.id)
            listener_dict = listener.to_dict()
            listener_dict['operating_status'] = db_listener.operating_status
            repo.listener.update(db_api.get_session(), listener.id,
                                 **listener_dict)
        elif create:
            repo.listener.update(db_api.get_session(), listener.id,
                                 operating_status=constants.ONLINE,
                                 provisioning_status=constants.ACTIVE)
        repo.load_balancer.update(db_api.get_session(),
                                  listener.load_balancer.id,
                                  operating_status=constants.ONLINE,
                                  provisioning_status=constants.ACTIVE)
        LOG.info("Simulated Controller Handler Thread Complete")

    def loadbalancer_controller(loadbalancer, delete=False, update=False,
                                create=False, failover=False):
        time.sleep(ASYNC_TIME)
        LOG.info("Simulating controller operation for loadbalancer...")

        if delete:
            repo.load_balancer.update(
                db_api.get_session(), id=loadbalancer.id,
                operating_status=constants.OFFLINE,
                provisioning_status=constants.DELETED)
        elif update:
            db_lb = repo.listener.get(db_api.get_session(), id=loadbalancer.id)
            lb_dict = loadbalancer.to_dict()
            lb_dict['operating_status'] = db_lb.operating_status
            repo.load_balancer.update(db_api.get_session(), loadbalancer.id,
                                      **lb_dict)
        elif create:
            repo.load_balancer.update(db_api.get_session(), id=loadbalancer.id,
                                      operating_status=constants.ONLINE,
                                      provisioning_status=constants.ACTIVE)
        elif failover:
            repo.load_balancer.update(
                db_api.get_session(), id=loadbalancer.id,
                operating_status=constants.ONLINE,
                provisioning_status=constants.PENDING_UPDATE)
        LOG.info("Simulated Controller Handler Thread Complete")

    controller = loadbalancer_controller
    if isinstance(data_model, data_models.Member):
        controller = member_controller
    elif isinstance(data_model, data_models.HealthMonitor):
        controller = health_monitor_controller
    elif isinstance(data_model, data_models.Pool):
        controller = pool_controller
    elif isinstance(data_model, data_models.Listener):
        controller = listener_controller

    thread = threading.Thread(target=controller, args=(data_model, delete,
                                                       update, create))
    thread.start()


class InvalidHandlerInputObject(Exception):
    message = "Invalid Input Object %(obj_type)s"

    def __init__(self, **kwargs):
        message = self.message % kwargs
        super(InvalidHandlerInputObject, self).__init__(message=message)


class LoadBalancerHandler(abstract_handler.BaseObjectHandler):

    def create(self, load_balancer_id):
        LOG.info("%(entity)s handling the creation of load balancer %(id)s",
                 {"entity": self.__class__.__name__, "id": load_balancer_id})
        simulate_controller(load_balancer_id, create=True)

    def update(self, old_lb, load_balancer):
        validate_input(data_models.LoadBalancer, load_balancer)
        LOG.info("%(entity)s handling the update of load balancer %(id)s",
                 {"entity": self.__class__.__name__, "id": old_lb.id})
        load_balancer.id = old_lb.id
        simulate_controller(load_balancer, update=True)

    def delete(self, load_balancer_id):
        LOG.info("%(entity)s handling the deletion of load balancer %(id)s",
                 {"entity": self.__class__.__name__, "id": load_balancer_id})
        simulate_controller(load_balancer_id, delete=True)


class ListenerHandler(abstract_handler.BaseObjectHandler):

    def create(self, listener_id):
        LOG.info("%(entity)s handling the creation of listener %(id)s",
                 {"entity": self.__class__.__name__, "id": listener_id})
        simulate_controller(listener_id, create=True)

    def update(self, old_listener, listener):
        validate_input(data_models.Listener, listener)
        LOG.info("%(entity)s handling the update of listener %(id)s",
                 {"entity": self.__class__.__name__, "id": old_listener.id})
        listener.id = old_listener.id
        simulate_controller(listener, update=True)

    def delete(self, listener_id):
        LOG.info("%(entity)s handling the deletion of listener %(id)s",
                 {"entity": self.__class__.__name__, "id": listener_id})
        simulate_controller(listener_id, delete=True)


class PoolHandler(abstract_handler.BaseObjectHandler):

    def create(self, pool_id):
        LOG.info("%(entity)s handling the creation of pool %(id)s",
                 {"entity": self.__class__.__name__, "id": pool_id})
        simulate_controller(pool_id, create=True)

    def update(self, old_pool, pool):
        validate_input(data_models.Pool, pool)
        LOG.info("%(entity)s handling the update of pool %(id)s",
                 {"entity": self.__class__.__name__, "id": old_pool.id})
        pool.id = old_pool.id
        simulate_controller(pool, update=True)

    def delete(self, pool_id):
        LOG.info("%(entity)s handling the deletion of pool %(id)s",
                 {"entity": self.__class__.__name__, "id": pool_id})
        simulate_controller(pool_id, delete=True)


class HealthMonitorHandler(abstract_handler.BaseObjectHandler):

    def create(self, pool_id):
        LOG.info("%(entity)s handling the creation of health monitor "
                 "on pool  %(id)s",
                 {"entity": self.__class__.__name__, "id": pool_id})
        simulate_controller(pool_id, create=True)

    def update(self, old_health_monitor, health_monitor):
        validate_input(data_models.HealthMonitor, health_monitor)
        LOG.info("%(entity)s handling the update of health monitor "
                 "on pool %(id)s",
                 {"entity": self.__class__.__name__,
                  "id": old_health_monitor.pool_id})
        health_monitor.pool_id = old_health_monitor.pool_id
        simulate_controller(health_monitor, update=True)

    def delete(self, pool_id):
        LOG.info("%(entity)s handling the deletion of health monitor "
                 "on pool %(id)s",
                 {"entity": self.__class__.__name__, "id": pool_id})
        simulate_controller(pool_id, delete=True)


class MemberHandler(abstract_handler.BaseObjectHandler):

    def create(self, member_id):
        LOG.info("%(entity)s handling the creation of member %(id)s",
                 {"entity": self.__class__.__name__, "id": member_id})
        simulate_controller(member_id, create=True)

    def update(self, old_member, member):
        validate_input(data_models.Member, member)
        LOG.info("%(entity)s handling the update of member %(id)s",
                 {"entity": self.__class__.__name__, "id": old_member.id})
        member.id = old_member.id
        simulate_controller(member, update=True)

    def batch_update(self, old_member_ids, new_member_ids, updated_members):
        for m in updated_members:
            validate_input(data_models.Member, m)
        LOG.info("%(entity)s handling the batch update of members: "
                 "old=%(old)s, new=%(new)s",
                 {"entity": self.__class__.__name__, "old": old_member_ids,
                  "new": new_member_ids})

        repo = repos.Repositories()
        old_members = [repo.member.get(db_api.get_session(), id=mid)
                       for mid in old_member_ids]
        new_members = [repo.member.get(db_api.get_session(), id=mid)
                       for mid in new_member_ids]
        all_members = []
        all_members.extend(old_members)
        all_members.extend(new_members)
        all_members.extend(updated_members)
        simulate_controller(all_members, batch_update=True)

    def delete(self, member_id):
        LOG.info("%(entity)s handling the deletion of member %(id)s",
                 {"entity": self.__class__.__name__, "id": member_id})
        simulate_controller(member_id, delete=True)


class L7PolicyHandler(abstract_handler.BaseObjectHandler):

    def create(self, l7policy_id):
        LOG.info("%(entity)s handling the creation of l7policy %(id)s",
                 {"entity": self.__class__.__name__, "id": l7policy_id})
        simulate_controller(l7policy_id, create=True)

    def update(self, old_l7policy, l7policy):
        validate_input(data_models.L7Policy, l7policy)
        LOG.info("%(entity)s handling the update of l7policy %(id)s",
                 {"entity": self.__class__.__name__, "id": old_l7policy.id})
        l7policy.id = old_l7policy.id
        simulate_controller(l7policy, update=True)

    def delete(self, l7policy_id):
        LOG.info("%(entity)s handling the deletion of l7policy %(id)s",
                 {"entity": self.__class__.__name__, "id": l7policy_id})
        simulate_controller(l7policy_id, delete=True)


class L7RuleHandler(abstract_handler.BaseObjectHandler):

    def create(self, l7rule):
        LOG.info("%(entity)s handling the creation of l7rule %(id)s",
                 {"entity": self.__class__.__name__, "id": l7rule.id})
        simulate_controller(l7rule, create=True)

    def update(self, old_l7rule, l7rule):
        validate_input(data_models.L7Rule, l7rule)
        LOG.info("%(entity)s handling the update of l7rule %(id)s",
                 {"entity": self.__class__.__name__, "id": old_l7rule.id})
        l7rule.id = old_l7rule.id
        simulate_controller(l7rule, update=True)

    def delete(self, l7rule):
        LOG.info("%(entity)s handling the deletion of l7rule %(id)s",
                 {"entity": self.__class__.__name__, "id": l7rule.id})
        simulate_controller(l7rule, delete=True)


class SimulatedControllerHandler(abstract_handler.BaseHandler):
    """Handler that simulates database calls of a successful controller."""
    load_balancer = LoadBalancerHandler()
    listener = ListenerHandler()
    pool = PoolHandler()
    health_monitor = HealthMonitorHandler()
    member = MemberHandler()
    l7policy = L7PolicyHandler()
    l7rule = L7RuleHandler()



#    Copyright 2016 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from octavia.common import constants
from octavia.tests.functional.api.v1 import base


class TestL7Policy(base.BaseAPITest):

    def setUp(self):
        super(TestL7Policy, self).setUp()
        self.lb = self.create_load_balancer(
            {'subnet_id': uuidutils.generate_uuid()})
        self.set_lb_status(self.lb.get('id'))
        self.listener = self.create_listener(self.lb.get('id'),
                                             constants.PROTOCOL_HTTP, 80)
        self.set_lb_status(self.lb.get('id'))
        self.pool = self.create_pool_sans_listener(
            self.lb.get('id'),
            constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'))
        self.l7policies_path = self.L7POLICIES_PATH.format(
            lb_id=self.lb.get('id'), listener_id=self.listener.get('id'))
        self.l7policy_path = self.l7policies_path + '/{l7policy_id}'

    def test_get(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        response = self.get(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')))
        response_body = response.json
        self.assertEqual(api_l7policy, response_body)

    def test_bad_get(self):
        self.get(self.l7policy_path.format(
            l7policy_id=uuidutils.generate_uuid()), status=404)

    def test_get_all(self):
        api_l7p_a = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        api_l7p_c = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        api_l7p_b = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT, position=2)
        self.set_lb_status(self.lb.get('id'))
        # api_l7p_b was inserted before api_l7p_c
        api_l7p_c['position'] = 3
        response = self.get(self.l7policies_path)
        response_body = response.json
        self.assertIsInstance(response_body, list)
        self.assertEqual(3, len(response_body))
        self.assertEqual(api_l7p_a, response_body[0])
        self.assertEqual(api_l7p_b, response_body[1])
        self.assertEqual(api_l7p_c, response_body[2])

    def test_empty_get_all(self):
        response = self.get(self.l7policies_path)
        response_body = response.json
        self.assertIsInstance(response_body, list)
        self.assertEqual(0, len(response_body))

    def test_create_reject_policy(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.assertEqual(constants.L7POLICY_ACTION_REJECT,
                         api_l7policy.get('action'))
        self.assertEqual(1, api_l7policy.get('position'))
        self.assertIsNone(api_l7policy.get('redirect_pool_id'))
        self.assertIsNone(api_l7policy.get('redirect_url'))
        self.assertTrue(api_l7policy.get('enabled'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_redirect_to_pool(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REDIRECT_TO_POOL,
            redirect_pool_id=self.pool.get('id'))
        self.assertEqual(constants.L7POLICY_ACTION_REDIRECT_TO_POOL,
                         api_l7policy.get('action'))
        self.assertEqual(1, api_l7policy.get('position'))
        self.assertEqual(self.pool.get('id'),
                         api_l7policy.get('redirect_pool_id'))
        self.assertIsNone(api_l7policy.get('redirect_url'))
        self.assertTrue(api_l7policy.get('enabled'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_redirect_to_url(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REDIRECT_TO_URL,
            redirect_url='http://www.example.com')
        self.assertEqual(constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                         api_l7policy.get('action'))
        self.assertEqual(1, api_l7policy.get('position'))
        self.assertIsNone(api_l7policy.get('redirect_pool_id'))
        self.assertEqual('http://www.example.com',
                         api_l7policy.get('redirect_url'))
        self.assertTrue(api_l7policy.get('enabled'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_with_id(self):
        l7p_id = uuidutils.generate_uuid()
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT, id=l7p_id)
        self.assertEqual(l7p_id, api_l7policy.get('id'))

    def test_create_with_duplicate_id(self):
        l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'), constants.ACTIVE)
        path = self.L7POLICIES_PATH.format(lb_id=self.lb.get('id'),
                                           listener_id=self.listener.get('id'))
        body = {'id': l7policy.get('id'),
                'action': constants.L7POLICY_ACTION_REJECT}
        self.post(path, body, status=409)

    def test_bad_create(self):
        l7policy = {'name': 'test1'}
        self.post(self.l7policies_path, l7policy, status=400)

    def test_bad_create_redirect_to_pool(self):
        l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_POOL,
                    'redirect_pool_id': uuidutils.generate_uuid()}
        self.post(self.l7policies_path, l7policy, status=404)

    def test_bad_create_redirect_to_url(self):
        l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                    'redirect_url': 'bad url'}
        self.post(self.l7policies_path, l7policy, status=400)

    def test_create_with_bad_handler(self):
        self.handler_mock().l7policy.create.side_effect = Exception()
        self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ERROR)

    def test_update(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'http://www.example.com'}
        response = self.put(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')), new_l7policy, status=202)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        response_body = response.json
        self.assertEqual(constants.L7POLICY_ACTION_REJECT,
                         response_body.get('action'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_bad_update(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        new_l7policy = {'action': 'bad action'}
        self.put(self.l7policy_path.format(l7policy_id=api_l7policy.get('id')),
                 new_l7policy, status=400)

    def test_bad_update_redirect_to_pool(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_POOL,
                        'redirect_pool_id': uuidutils.generate_uuid()}
        self.put(self.l7policy_path.format(l7policy_id=api_l7policy.get('id')),
                 new_l7policy, status=404)

    def test_bad_update_redirect_to_url(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'bad url'}
        self.put(self.l7policy_path.format(l7policy_id=api_l7policy.get('id')),
                 new_l7policy, status=400)

    def test_update_with_bad_handler(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'http://www.example.com'}
        self.handler_mock().l7policy.update.side_effect = Exception()
        self.put(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')), new_l7policy, status=202)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ERROR)

    def test_update_redirect_to_pool_bad_pool_id(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        path = self.l7policy_path.format(l7policy_id=api_l7policy.get('id'))
        new_l7policy = {'redirect_pool_id': uuidutils.generate_uuid()}
        self.put(path, new_l7policy, status=404)

    def test_update_redirect_to_pool_minimal(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        path = self.l7policy_path.format(l7policy_id=api_l7policy.get('id'))
        new_l7policy = {'redirect_pool_id': self.pool.get('id')}
        self.put(path, new_l7policy, status=202)

    def test_update_redirect_to_url_bad_url(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        path = self.l7policy_path.format(l7policy_id=api_l7policy.get('id'))
        new_l7policy = {'redirect_url': 'bad-url'}
        self.put(path, new_l7policy, status=400)

    def test_update_redirect_to_url_minimal(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        path = self.l7policy_path.format(l7policy_id=api_l7policy.get('id'))
        new_l7policy = {'redirect_url': 'http://www.example.com/'}
        self.put(path, new_l7policy, status=202)

    def test_delete(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        response = self.get(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')))
        self.assertEqual(api_l7policy, response.json)
        self.delete(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_bad_delete(self):
        self.delete(self.l7policy_path.format(
            l7policy_id=uuidutils.generate_uuid()), status=404)

    def test_delete_with_bad_handler(self):
        api_l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        response = self.get(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')))
        self.assertEqual(api_l7policy, response.json)
        self.handler_mock().l7policy.delete.side_effect = Exception()
        self.delete(self.l7policy_path.format(
            l7policy_id=api_l7policy.get('id')))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ERROR)

    def test_create_when_lb_pending_update(self):
        self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        self.put(self.LB_PATH.format(lb_id=self.lb.get('id')),
                 body={'name': 'test_name_change'})
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'http://www.example.com'}
        self.post(self.l7policies_path, body=new_l7policy, status=409)

    def test_update_when_lb_pending_update(self):
        l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        self.put(self.LB_PATH.format(lb_id=self.lb.get('id')),
                 body={'name': 'test_name_change'})
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'http://www.example.com'}
        self.put(self.l7policy_path.format(l7policy_id=l7policy.get('id')),
                 body=new_l7policy, status=409)

    def test_delete_when_lb_pending_update(self):
        l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        self.put(self.LB_PATH.format(lb_id=self.lb.get('id')),
                 body={'name': 'test_name_change'})
        self.delete(self.l7policy_path.format(l7policy_id=l7policy.get('id')),
                    status=409)

    def test_create_when_lb_pending_delete(self):
        self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        self.delete(self.LB_DELETE_CASCADE_PATH.format(
            lb_id=self.lb.get('id')))
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'http://www.example.com'}
        self.post(self.l7policies_path, body=new_l7policy, status=409)

    def test_update_when_lb_pending_delete(self):
        l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        self.delete(self.LB_DELETE_CASCADE_PATH.format(
            lb_id=self.lb.get('id')))
        new_l7policy = {'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                        'redirect_url': 'http://www.example.com'}
        self.put(self.l7policy_path.format(l7policy_id=l7policy.get('id')),
                 body=new_l7policy, status=409)

    def test_delete_when_lb_pending_delete(self):
        l7policy = self.create_l7policy(
            self.lb.get('id'), self.listener.get('id'),
            constants.L7POLICY_ACTION_REJECT)
        self.set_lb_status(self.lb.get('id'))
        self.delete(self.LB_DELETE_CASCADE_PATH.format(
            lb_id=self.lb.get('id')))
        self.delete(self.l7policy_path.format(l7policy_id=l7policy.get('id')),
                    status=409)



#    Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutronclient.common import exceptions as neutron_client_exceptions

from octavia.common import clients
from octavia.common import data_models
from octavia.network import base as network_base
from octavia.network import data_models as network_models
from octavia.network.drivers.neutron import base as neutron_base
from octavia.network.drivers.neutron import utils
from octavia.tests.common import constants as t_constants
from octavia.tests.common import data_model_helpers as dmh
from octavia.tests.unit import base


class TestBaseNeutronNetworkDriver(base.TestCase):

    def _instantiate_partial_abc(self, abclass):
        if "__abstractmethods__" not in abclass.__dict__:
            return abclass()
        new_dict = abclass.__dict__.copy()
        for abstractmethod in abclass.__abstractmethods__:
            new_dict[abstractmethod] = lambda x, *args, **kw: (x, args, kw)
        impl_class = type("partially_implemented_abc_%s" % abclass.__name__,
                          (abclass,), new_dict)
        return impl_class()

    def setUp(self):
        super(TestBaseNeutronNetworkDriver, self).setUp()
        with mock.patch('octavia.common.clients.neutron_client.Client',
                        autospec=True) as neutron_client:
            client = neutron_client(clients.NEUTRON_VERSION)
            client.list_extensions.return_value = {
                'extensions': [
                    {'alias': neutron_base.SEC_GRP_EXT_ALIAS}
                ]
            }
            self.k_session = mock.patch(
                'keystoneauth1.session.Session').start()
            self.driver = self._instantiate_partial_abc(
                neutron_base.BaseNeutronDriver)

    def test__check_extension_enabled(self):
        show_extension = self.driver.neutron_client.show_extension
        show_extension.side_effect = [None, neutron_client_exceptions.NotFound]

        self.assertTrue(self.driver._check_extension_enabled('TEST1'))
        self.assertFalse(self.driver._check_extension_enabled('TEST2'))
        show_extension.assert_has_calls(
            [mock.call('TEST1'), mock.call('TEST2')])

    def test__check_extension_enabled_cached(self):
        show_extension = self.driver.neutron_client.show_extension

        self.driver._check_extension_cache = {'TEST1': True, 'TEST2': False}
        self.assertTrue(self.driver._check_extension_enabled('TEST1'))
        self.assertFalse(self.driver._check_extension_enabled('TEST2'))
        self.assertNotIn(mock.call('TEST1'), show_extension.mock_calls)
        self.assertNotIn(mock.call('TEST2'), show_extension.mock_calls)

    def test__add_allowed_address_pair_to_port(self):
        self.driver._add_allowed_address_pair_to_port(
            t_constants.MOCK_PORT_ID, t_constants.MOCK_IP_ADDRESS)
        expected_aap_dict = {
            'port': {
                'allowed_address_pairs': [
                    {'ip_address': t_constants.MOCK_IP_ADDRESS}]}}
        self.driver.neutron_client.update_port.assert_has_calls([
            mock.call(t_constants.MOCK_PORT_ID, expected_aap_dict)])

    def test__add_security_group_to_port(self):
        self.driver._add_security_group_to_port(
            t_constants.MOCK_SECURITY_GROUP_ID, t_constants.MOCK_PORT_ID)
        expected_sg_dict = {
            'port': {
                'security_groups': [
                    t_constants.MOCK_SECURITY_GROUP_ID]}}
        self.driver.neutron_client.update_port.assert_has_calls([
            mock.call(t_constants.MOCK_PORT_ID, expected_sg_dict)])

    def test__add_security_group_to_port_with_port_not_found(self):
        self.driver.neutron_client.update_port.side_effect = (
            neutron_client_exceptions.PortNotFoundClient)
        self.assertRaises(
            network_base.PortNotFound,
            self.driver._add_security_group_to_port,
            t_constants.MOCK_SECURITY_GROUP_ID, t_constants.MOCK_PORT_ID)

    def test__add_security_group_to_port_with_other_exception(self):
        self.driver.neutron_client.update_port.side_effect = IOError
        self.assertRaises(
            network_base.NetworkException,
            self.driver._add_security_group_to_port,
            t_constants.MOCK_SECURITY_GROUP_ID, t_constants.MOCK_PORT_ID)

    def test__get_ports_by_security_group(self):
        self.driver.neutron_client.list_ports.return_value = {
            "ports": [
                t_constants.MOCK_NEUTRON_PORT['port'],
                t_constants.MOCK_NEUTRON_PORT2['port']]
        }
        ports = self.driver._get_ports_by_security_group(
            t_constants.MOCK_SECURITY_GROUP_ID)
        self.assertEqual(1, len(ports))
        self.assertIn(t_constants.MOCK_NEUTRON_PORT['port'], ports)

    def test__create_security_group(self):
        sg_return = self.driver._create_security_group(
            t_constants.MOCK_SECURITY_GROUP_NAME)
        expected_sec_grp_dict = {
            'security_group': {
                'name': t_constants.MOCK_SECURITY_GROUP_NAME}}
        self.driver.neutron_client.create_security_group.assert_has_calls([
            mock.call(expected_sec_grp_dict)])
        self.assertEqual(
            sg_return,
            self.driver.neutron_client.create_security_group()[
                'security_group'])

    def test__create_security_group_rule(self):
        self.driver._create_security_group_rule(
            sec_grp_id=t_constants.MOCK_SECURITY_GROUP_ID,
            direction=1,
            protocol=2,
            port_min=3,
            port_max=4,
            ethertype=5)
        expected_sec_grp_rule_dict = {
            'security_group_rule': {
                'security_group_id': t_constants.MOCK_SECURITY_GROUP_ID,
                'direction': 1,
                'protocol': 2,
                'port_range_min': 3,
                'port_range_max': 4,
                'ethertype': 5}}
        self.driver.neutron_client.create_security_group_rule.assert_has_calls(
            [mock.call(expected_sec_grp_rule_dict)])

    def test__port_to_vip(self):
        lb = dmh.generate_load_balancer_tree()
        lb.vip.subnet_id = t_constants.MOCK_SUBNET_ID
        port = utils.convert_port_dict_to_model(t_constants.MOCK_NEUTRON_PORT)
        vip = self.driver._port_to_vip(port, lb)
        self.assertIsInstance(vip, data_models.Vip)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS, vip.ip_address)
        self.assertEqual(t_constants.MOCK_SUBNET_ID, vip.subnet_id)
        self.assertEqual(t_constants.MOCK_PORT_ID, vip.port_id)
        self.assertEqual(lb.id, vip.load_balancer_id)

    def test__nova_interface_to_octavia_interface(self):
        nova_interface = t_constants.MockNovaInterface()
        nova_interface.net_id = '1'
        nova_interface.port_id = '2'
        nova_interface.fixed_ips = [{'ip_address': '10.0.0.1'}]
        interface = self.driver._nova_interface_to_octavia_interface(
            '3', nova_interface)
        self.assertEqual('1', interface.network_id)
        self.assertEqual('2', interface.port_id)
        ips = [fixed_ip.ip_address for fixed_ip in interface.fixed_ips]
        self.assertIn('10.0.0.1', ips)

    def test_get_plugged_networks(self):
        list_ports = self.driver.neutron_client.list_ports
        list_ports.side_effect = TypeError
        o_ifaces = self.driver.get_plugged_networks(
            t_constants.MOCK_DEVICE_ID)
        self.assertEqual(0, len(o_ifaces))
        list_ports.side_effect = None
        list_ports.reset_mock()
        port1 = t_constants.MOCK_NEUTRON_PORT['port']
        port2 = {
            'id': '4', 'network_id': '3', 'fixed_ips':
            [{'ip_address': '10.0.0.2'}]
        }
        list_ports.return_value = {'ports': [port1, port2]}
        plugged_networks = self.driver.get_plugged_networks(
            t_constants.MOCK_DEVICE_ID)
        for pn in plugged_networks:
            self.assertIn(pn.port_id, [port1.get('id'), port2.get('id')])
            self.assertIn(pn.network_id, [port1.get('network_id'),
                                          port2.get('network_id')])
            for fixed_ip in pn.fixed_ips:
                self.assertIn(fixed_ip.ip_address,
                              [port1['fixed_ips'][0]['ip_address'],
                               port2['fixed_ips'][0]['ip_address']])

    def test_get_network(self):
        show_network = self.driver.neutron_client.show_network
        show_network.return_value = {'network': {
            'id': t_constants.MOCK_NETWORK_ID,
            'subnets': [t_constants.MOCK_SUBNET_ID]}}
        network = self.driver.get_network(t_constants.MOCK_NETWORK_ID)
        self.assertIsInstance(network, network_models.Network)
        self.assertEqual(t_constants.MOCK_NETWORK_ID, network.id)
        self.assertEqual(1, len(network.subnets))
        self.assertEqual(t_constants.MOCK_SUBNET_ID, network.subnets[0])

    def test_get_subnet(self):
        show_subnet = self.driver.neutron_client.show_subnet
        show_subnet.return_value = {'subnet': {
            'id': t_constants.MOCK_SUBNET_ID,
            'gateway_ip': t_constants.MOCK_IP_ADDRESS,
            'cidr': t_constants.MOCK_CIDR}}
        subnet = self.driver.get_subnet(t_constants.MOCK_SUBNET_ID)
        self.assertIsInstance(subnet, network_models.Subnet)
        self.assertEqual(t_constants.MOCK_SUBNET_ID, subnet.id)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS, subnet.gateway_ip)
        self.assertEqual(t_constants.MOCK_CIDR, subnet.cidr)

    def test_get_port(self):
        show_port = self.driver.neutron_client.show_port
        show_port.return_value = {'port': {
            'id': t_constants.MOCK_PORT_ID,
            'mac_address': t_constants.MOCK_MAC_ADDR,
            'network_id': t_constants.MOCK_NETWORK_ID,
            'fixed_ips': [{
                'subnet_id': t_constants.MOCK_SUBNET_ID,
                'ip_address': t_constants.MOCK_IP_ADDRESS
            }]}}
        port = self.driver.get_port(t_constants.MOCK_PORT_ID)
        self.assertIsInstance(port, network_models.Port)
        self.assertEqual(t_constants.MOCK_PORT_ID, port.id)
        self.assertEqual(t_constants.MOCK_MAC_ADDR, port.mac_address)
        self.assertEqual(t_constants.MOCK_NETWORK_ID, port.network_id)
        self.assertEqual(1, len(port.fixed_ips))
        self.assertIsInstance(port.fixed_ips[0], network_models.FixedIP)
        self.assertEqual(t_constants.MOCK_SUBNET_ID,
                         port.fixed_ips[0].subnet_id)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS,
                         port.fixed_ips[0].ip_address)

    def test_get_network_by_name(self):
        list_network = self.driver.neutron_client.list_networks
        list_network.return_value = {'networks': [{'network': {
            'id': t_constants.MOCK_NETWORK_ID,
            'name': t_constants.MOCK_NETWORK_NAME,
            'subnets': [t_constants.MOCK_SUBNET_ID]}}]}
        network = self.driver.get_network_by_name(
            t_constants.MOCK_NETWORK_NAME)
        self.assertIsInstance(network, network_models.Network)
        self.assertEqual(t_constants.MOCK_NETWORK_ID, network.id)
        self.assertEqual(t_constants.MOCK_NETWORK_NAME, network.name)
        self.assertEqual(1, len(network.subnets))
        self.assertEqual(t_constants.MOCK_SUBNET_ID, network.subnets[0])
        # Negative
        list_network.side_effect = neutron_client_exceptions.NotFound
        self.assertRaises(network_base.NetworkNotFound,
                          self.driver.get_network_by_name,
                          t_constants.MOCK_NETWORK_NAME)
        list_network.side_effect = Exception
        self.assertRaises(network_base.NetworkException,
                          self.driver.get_network_by_name,
                          t_constants.MOCK_NETWORK_NAME)

    def test_get_subnet_by_name(self):
        list_subnet = self.driver.neutron_client.list_subnets
        list_subnet.return_value = {'subnets': [{'subnet': {
            'id': t_constants.MOCK_SUBNET_ID,
            'name': t_constants.MOCK_SUBNET_NAME,
            'gateway_ip': t_constants.MOCK_IP_ADDRESS,
            'cidr': t_constants.MOCK_CIDR}}]}
        subnet = self.driver.get_subnet_by_name(t_constants.MOCK_SUBNET_NAME)
        self.assertIsInstance(subnet, network_models.Subnet)
        self.assertEqual(t_constants.MOCK_SUBNET_ID, subnet.id)
        self.assertEqual(t_constants.MOCK_SUBNET_NAME, subnet.name)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS, subnet.gateway_ip)
        self.assertEqual(t_constants.MOCK_CIDR, subnet.cidr)
        # Negative
        list_subnet.side_effect = neutron_client_exceptions.NotFound
        self.assertRaises(network_base.SubnetNotFound,
                          self.driver.get_subnet_by_name,
                          t_constants.MOCK_SUBNET_NAME)
        list_subnet.side_effect = Exception
        self.assertRaises(network_base.NetworkException,
                          self.driver.get_subnet_by_name,
                          t_constants.MOCK_SUBNET_NAME)

    def test_get_port_by_name(self):
        list_port = self.driver.neutron_client.list_ports
        list_port.return_value = {'ports': [{'port': {
            'id': t_constants.MOCK_PORT_ID,
            'name': t_constants.MOCK_PORT_NAME,
            'mac_address': t_constants.MOCK_MAC_ADDR,
            'network_id': t_constants.MOCK_NETWORK_ID,
            'fixed_ips': [{
                'subnet_id': t_constants.MOCK_SUBNET_ID,
                'ip_address': t_constants.MOCK_IP_ADDRESS
            }]}}]}
        port = self.driver.get_port_by_name(t_constants.MOCK_PORT_NAME)
        self.assertIsInstance(port, network_models.Port)
        self.assertEqual(t_constants.MOCK_PORT_ID, port.id)
        self.assertEqual(t_constants.MOCK_PORT_NAME, port.name)
        self.assertEqual(t_constants.MOCK_MAC_ADDR, port.mac_address)
        self.assertEqual(t_constants.MOCK_NETWORK_ID, port.network_id)
        self.assertEqual(1, len(port.fixed_ips))
        self.assertIsInstance(port.fixed_ips[0], network_models.FixedIP)
        self.assertEqual(t_constants.MOCK_SUBNET_ID,
                         port.fixed_ips[0].subnet_id)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS,
                         port.fixed_ips[0].ip_address)
        # Negative
        list_port.side_effect = neutron_client_exceptions.NotFound
        self.assertRaises(network_base.PortNotFound,
                          self.driver.get_port_by_name,
                          t_constants.MOCK_PORT_NAME)
        list_port.side_effect = Exception
        self.assertRaises(network_base.NetworkException,
                          self.driver.get_port_by_name,
                          t_constants.MOCK_PORT_NAME)

    def test_get_port_by_net_id_device_id(self):
        list_port = self.driver.neutron_client.list_ports
        list_port.return_value = {'ports': [{'port': {
            'id': t_constants.MOCK_PORT_ID,
            'name': t_constants.MOCK_PORT_NAME,
            'mac_address': t_constants.MOCK_MAC_ADDR,
            'network_id': t_constants.MOCK_NETWORK_ID,
            'device_id': t_constants.MOCK_DEVICE_ID,
            'fixed_ips': [{
                'subnet_id': t_constants.MOCK_SUBNET_ID,
                'ip_address': t_constants.MOCK_IP_ADDRESS
            }]}}]}
        port = self.driver.get_port_by_net_id_device_id(
            t_constants.MOCK_NETWORK_ID, t_constants.MOCK_DEVICE_ID)
        self.assertIsInstance(port, network_models.Port)
        self.assertEqual(t_constants.MOCK_PORT_ID, port.id)
        self.assertEqual(t_constants.MOCK_DEVICE_ID, port.device_id)
        self.assertEqual(t_constants.MOCK_PORT_NAME, port.name)
        self.assertEqual(t_constants.MOCK_MAC_ADDR, port.mac_address)
        self.assertEqual(t_constants.MOCK_NETWORK_ID, port.network_id)
        self.assertEqual(1, len(port.fixed_ips))
        self.assertIsInstance(port.fixed_ips[0], network_models.FixedIP)
        self.assertEqual(t_constants.MOCK_SUBNET_ID,
                         port.fixed_ips[0].subnet_id)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS,
                         port.fixed_ips[0].ip_address)
        # Negative
        list_port.side_effect = neutron_client_exceptions.NotFound
        self.assertRaises(network_base.PortNotFound,
                          self.driver.get_port_by_net_id_device_id,
                          t_constants.MOCK_PORT_NAME,
                          t_constants.MOCK_DEVICE_ID)
        list_port.side_effect = Exception
        self.assertRaises(network_base.NetworkException,
                          self.driver.get_port_by_net_id_device_id,
                          t_constants.MOCK_NETWORK_ID,
                          t_constants.MOCK_DEVICE_ID)

    def test_get_ports_by_net_id_device_id(self):
        """Test get_port_by_net_id_device_id, when port is not unique.

        The expected result is: only the first port is returned.
        """

        list_port = self.driver.neutron_client.list_ports
        list_port.return_value = {
            'ports': [t_constants.MOCK_NEUTRON_PORT,
                      t_constants.MOCK_NEUTRON_PORT2,
                      ],
        }

        port = self.driver.get_port_by_net_id_device_id(
            t_constants.MOCK_NETWORK_ID, t_constants.MOCK_DEVICE_ID)
        self.assertIsInstance(port, network_models.Port)
        self.assertEqual(t_constants.MOCK_PORT_ID, port.id)
        self.assertEqual(t_constants.MOCK_DEVICE_ID, port.device_id)
        self.assertEqual(t_constants.MOCK_PORT_NAME, port.name)
        self.assertEqual(t_constants.MOCK_MAC_ADDR, port.mac_address)
        self.assertEqual(t_constants.MOCK_NETWORK_ID, port.network_id)
        self.assertEqual(1, len(port.fixed_ips))
        self.assertIsInstance(port.fixed_ips[0], network_models.FixedIP)
        self.assertEqual(t_constants.MOCK_SUBNET_ID,
                         port.fixed_ips[0].subnet_id)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS,
                         port.fixed_ips[0].ip_address)
        # Negative
        list_port.side_effect = neutron_client_exceptions.NotFound
        self.assertRaises(network_base.PortNotFound,
                          self.driver.get_port_by_net_id_device_id,
                          t_constants.MOCK_PORT_NAME,
                          t_constants.MOCK_DEVICE_ID)
        list_port.side_effect = Exception
        self.assertRaises(network_base.NetworkException,
                          self.driver.get_port_by_net_id_device_id,
                          t_constants.MOCK_NETWORK_ID,
                          t_constants.MOCK_DEVICE_ID)

    def test_get_multiple_ports_by_net_id_device_id(self):
        """Test _get_resources_by_filters, when result is not unique"""
        list_port = self.driver.neutron_client.list_ports
        list_port.return_value = {
            'ports': [t_constants.MOCK_NEUTRON_PORT,
                      t_constants.MOCK_NEUTRON_PORT2,
                      ],
        }

        ports = self.driver._get_resources_by_filters(
            'port',
            network_id=t_constants.MOCK_NETWORK_ID,
            device_id=t_constants.MOCK_DEVICE_ID,
        )
        self.assertIsInstance(ports, list)
        port1, port2 = ports

        self.assertEqual(t_constants.MOCK_PORT_ID, port1.id)
        self.assertEqual(t_constants.MOCK_PORT_ID2, port2.id)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS,
                         port1.fixed_ips[0].ip_address)
        self.assertEqual(t_constants.MOCK_IP_ADDRESS2,
                         port2.fixed_ips[0].ip_address)

    def test_get_unique_port_by_name(self):
        """Test _get_resources_by_filters, when result is unique"""
        list_port = self.driver.neutron_client.list_ports
        list_port.return_value = {
            'ports': [t_constants.MOCK_NEUTRON_PORT]
        }

        port = self.driver._get_resources_by_filters(
            'port', unique_item=True, name=t_constants.MOCK_PORT_NAME)

        self.assertIsInstance(port, network_models.Port)
        self.assertEqual(t_constants.MOCK_PORT_ID, port.id)

    def test_get_non_existing_port_by_name(self):
        """Test _get_resources_by_filters, when result is empty"""
        list_port = self.driver.neutron_client.list_ports
        list_port.return_value = {'ports': []}

        self.assertRaises(network_base.PortNotFound,
                          self.driver._get_resources_by_filters,
                          'port', unique_item=True, name='port1')

    def test_get_qos_policy(self):
        get_qos = self.driver.neutron_client.show_qos_policy
        get_qos.return_value = {'policy': {
            'id': t_constants.MOCK_NEUTRON_QOS_POLICY_ID}}
        qos = self.driver.get_qos_policy(
            t_constants.MOCK_NEUTRON_QOS_POLICY_ID)
        self.assertIsInstance(qos, network_models.QosPolicy)
        self.assertEqual(t_constants.MOCK_NEUTRON_QOS_POLICY_ID,
                         qos.id)

        get_qos.side_effect = neutron_client_exceptions.NotFound
        self.assertRaises(network_base.QosPolicyNotFound,
                          self.driver.get_qos_policy,
                          t_constants.MOCK_NEUTRON_QOS_POLICY_ID)

        get_qos.side_effect = neutron_client_exceptions.ServiceUnavailable
        self.assertRaises(network_base.NetworkException,
                          self.driver.get_qos_policy,
                          t_constants.MOCK_NEUTRON_QOS_POLICY_ID)

    def test_apply_or_undo_qos_on_port(self):
        # The apply and undo qos function use the same "update_port" with
        # neutron client. So testing them in one Uts.
        update_port = self.driver.neutron_client.update_port
        update_port.side_effect = neutron_client_exceptions.PortNotFoundClient
        self.assertRaises(network_base.PortNotFound,
                          self.driver.apply_qos_on_port,
                          t_constants.MOCK_PORT_ID,
                          t_constants.MOCK_NEUTRON_QOS_POLICY_ID)

        update_port.side_effect = neutron_client_exceptions.ServiceUnavailable
        self.assertRaises(network_base.NetworkException,
                          self.driver.apply_qos_on_port,
                          t_constants.MOCK_PORT_ID,
                          t_constants.MOCK_NEUTRON_QOS_POLICY_ID)



#    Copyright (c) 2014 Rackspace
#    Copyright (c) 2016 Blue Box, an IBM Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import six
from sqlalchemy.orm import collections

from octavia.common import constants


class BaseDataModel(object):
    def to_dict(self, calling_classes=None, recurse=False, **kwargs):
        """Converts a data model to a dictionary."""
        calling_classes = calling_classes or []
        ret = {}
        for attr in self.__dict__:
            if attr.startswith('_') or not kwargs.get(attr, True):
                continue
            value = self.__dict__[attr]

            if recurse:
                if isinstance(getattr(self, attr), list):
                    ret[attr] = []
                    for item in value:
                        if isinstance(item, BaseDataModel):
                            if type(self) not in calling_classes:
                                ret[attr].append(
                                    item.to_dict(calling_classes=(
                                        calling_classes + [type(self)])))
                            else:
                                ret[attr] = None
                        else:
                            ret[attr] = item
                elif isinstance(getattr(self, attr), BaseDataModel):
                    if type(self) not in calling_classes:
                        ret[attr] = value.to_dict(
                            calling_classes=calling_classes + [type(self)])
                    else:
                        ret[attr] = None
                elif six.PY2 and isinstance(value, six.text_type):
                    ret[attr.encode('utf8')] = value.encode('utf8')
                else:
                    ret[attr] = value
            else:
                if isinstance(getattr(self, attr), (BaseDataModel, list)):
                    ret[attr] = None
                else:
                    ret[attr] = value

        return ret

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.to_dict() == other.to_dict()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_dict(cls, dict):
        return cls(**dict)

    @classmethod
    def _name(cls):
        """Returns class name in a more human readable form."""
        # Split the class name up by capitalized words
        return ' '.join(re.findall('[A-Z][^A-Z]*', cls.__name__))

    def _get_unique_key(self, obj=None):
        """Returns a unique key for passed object for data model building."""
        obj = obj or self
        # First handle all objects with their own ID, then handle subordinate
        # objects.
        if obj.__class__.__name__ in ['Member', 'Pool', 'LoadBalancer',
                                      'Listener', 'Amphora', 'L7Policy',
                                      'L7Rule']:
            return obj.__class__.__name__ + obj.id
        elif obj.__class__.__name__ in ['SessionPersistence', 'HealthMonitor']:
            return obj.__class__.__name__ + obj.pool_id
        elif obj.__class__.__name__ in ['ListenerStatistics']:
            return obj.__class__.__name__ + obj.listener_id + obj.amphora_id
        elif obj.__class__.__name__ in ['VRRPGroup', 'Vip']:
            return obj.__class__.__name__ + obj.load_balancer_id
        elif obj.__class__.__name__ in ['AmphoraHealth']:
            return obj.__class__.__name__ + obj.amphora_id
        elif obj.__class__.__name__ in ['SNI']:
            return (obj.__class__.__name__ +
                    obj.listener_id + obj.tls_container_id)
        else:
            raise NotImplementedError

    def _find_in_graph(self, key, _visited_nodes=None):
        """Locates an object with the given unique key in the current

        object graph and returns a reference to it.
        """
        _visited_nodes = _visited_nodes or []
        mykey = self._get_unique_key()
        if mykey in _visited_nodes:
            # Seen this node already, don't traverse further
            return None
        elif mykey == key:
            return self
        else:
            _visited_nodes.append(mykey)
            attr_names = [attr_name for attr_name in dir(self)
                          if not attr_name.startswith('_')]
            for attr_name in attr_names:
                attr = getattr(self, attr_name)
                if isinstance(attr, BaseDataModel):
                    result = attr._find_in_graph(
                        key, _visited_nodes=_visited_nodes)
                    if result is not None:
                        return result
                elif isinstance(attr, (collections.InstrumentedList, list)):
                    for item in attr:
                        if isinstance(item, BaseDataModel):
                            result = item._find_in_graph(
                                key, _visited_nodes=_visited_nodes)
                            if result is not None:
                                return result
        # If we are here we didn't find it.
        return None

    def update(self, update_dict):
        """Generic update method which works for simple,

        non-relational attributes.
        """
        for key, value in update_dict.items():
            setattr(self, key, value)


class SessionPersistence(BaseDataModel):

    def __init__(self, pool_id=None, type=None, cookie_name=None,
                 pool=None):
        self.pool_id = pool_id
        self.type = type
        self.cookie_name = cookie_name
        self.pool = pool

    def delete(self):
        self.pool.session_persistence = None


class ListenerStatistics(BaseDataModel):

    def __init__(self, listener_id=None, amphora_id=None, bytes_in=0,
                 bytes_out=0, active_connections=0,
                 total_connections=0, request_errors=0):
        self.listener_id = listener_id
        self.amphora_id = amphora_id
        self.bytes_in = bytes_in
        self.bytes_out = bytes_out
        self.active_connections = active_connections
        self.total_connections = total_connections
        self.request_errors = request_errors

    def get_stats(self):
        stats = {
            'bytes_in': self.bytes_in,
            'bytes_out': self.bytes_out,
            'active_connections': self.active_connections,
            'total_connections': self.total_connections,
            'request_errors': self.request_errors,
        }
        return stats

    def __iadd__(self, other):

        if isinstance(other, ListenerStatistics):
            self.bytes_in += other.bytes_in
            self.bytes_out += other.bytes_out
            self.request_errors += other.request_errors
            self.total_connections += other.total_connections

        return self


class LoadBalancerStatistics(BaseDataModel):

    def __init__(self, bytes_in=0, bytes_out=0, active_connections=0,
                 total_connections=0, request_errors=0, listeners=None):
        self.bytes_in = bytes_in
        self.bytes_out = bytes_out
        self.active_connections = active_connections
        self.total_connections = total_connections
        self.request_errors = request_errors
        self.listeners = listeners or []

    def get_stats(self):
        stats = {
            'bytes_in': self.bytes_in,
            'bytes_out': self.bytes_out,
            'active_connections': self.active_connections,
            'total_connections': self.total_connections,
            'request_errors': self.request_errors,
        }
        return stats


class HealthMonitor(BaseDataModel):

    def __init__(self, id=None, project_id=None, pool_id=None, type=None,
                 delay=None, timeout=None, fall_threshold=None,
                 rise_threshold=None, http_method=None, url_path=None,
                 expected_codes=None, enabled=None, pool=None, name=None,
                 provisioning_status=None, operating_status=None,
                 created_at=None, updated_at=None):
        self.id = id
        self.project_id = project_id
        self.pool_id = pool_id
        self.type = type
        self.delay = delay
        self.timeout = timeout
        self.fall_threshold = fall_threshold
        self.rise_threshold = rise_threshold
        self.http_method = http_method
        self.url_path = url_path
        self.expected_codes = expected_codes
        self.enabled = enabled
        self.pool = pool
        self.provisioning_status = provisioning_status
        self.operating_status = operating_status
        self.name = name
        self.created_at = created_at
        self.updated_at = updated_at

    def delete(self):
        self.pool.health_monitor = None


class Pool(BaseDataModel):
    def __init__(self, id=None, project_id=None, name=None, description=None,
                 protocol=None, lb_algorithm=None, enabled=None,
                 operating_status=None, members=None, health_monitor=None,
                 session_persistence=None, load_balancer_id=None,
                 load_balancer=None, listeners=None, l7policies=None,
                 created_at=None, updated_at=None, provisioning_status=None):
        self.id = id
        self.project_id = project_id
        self.name = name
        self.description = description
        self.load_balancer_id = load_balancer_id
        self.load_balancer = load_balancer
        self.protocol = protocol
        self.lb_algorithm = lb_algorithm
        self.enabled = enabled
        self.operating_status = operating_status
        self.members = members or []
        self.health_monitor = health_monitor
        self.session_persistence = session_persistence
        self.listeners = listeners or []
        self.l7policies = l7policies or []
        self.created_at = created_at
        self.updated_at = updated_at
        self.provisioning_status = provisioning_status

    def update(self, update_dict):
        for key, value in update_dict.items():
            if key == 'session_persistence':
                if value is None or value == {}:
                    if self.session_persistence is not None:
                        self.session_persistence.delete()
                elif self.session_persistence is not None:
                    self.session_persistence.update(value)
                else:
                    value.update({'pool_id': self.id})
                    self.session_persistence = SessionPersistence(**value)
            else:
                setattr(self, key, value)

    def delete(self):
        for listener in self.listeners:
            if listener.default_pool_id == self.id:
                listener.default_pool = None
                listener.default_pool_id = None
            for pool in listener.pools:
                if pool.id == self.id:
                    listener.pools.remove(pool)
                    break
        for pool in self.load_balancer.pools:
            if pool.id == self.id:
                self.load_balancer.pools.remove(pool)
                break
        for l7policy in self.l7policies:
            if l7policy.redirect_pool_id == self.id:
                # Technically this should never happen, as we block deletion
                # of pools in use by L7Policies at the API. However, we should
                # probably keep this here in case the data model gets
                # manipulated in some other way in the future.
                l7policy.action = constants.L7POLICY_ACTION_REJECT
                l7policy.redirect_pool = None
                l7policy.redirect_pool_id = None


class Member(BaseDataModel):

    def __init__(self, id=None, project_id=None, pool_id=None, ip_address=None,
                 protocol_port=None, weight=None, enabled=None,
                 subnet_id=None, operating_status=None, pool=None,
                 created_at=None, updated_at=None, provisioning_status=None,
                 name=None, monitor_address=None, monitor_port=None):
        self.id = id
        self.project_id = project_id
        self.pool_id = pool_id
        self.ip_address = ip_address
        self.protocol_port = protocol_port
        self.weight = weight
        self.enabled = enabled
        self.subnet_id = subnet_id
        self.operating_status = operating_status
        self.pool = pool
        self.created_at = created_at
        self.updated_at = updated_at
        self.provisioning_status = provisioning_status
        self.name = name
        self.monitor_address = monitor_address
        self.monitor_port = monitor_port

    def delete(self):
        for mem in self.pool.members:
            if mem.id == self.id:
                self.pool.members.remove(mem)
                break


class Listener(BaseDataModel):

    def __init__(self, id=None, project_id=None, name=None, description=None,
                 default_pool_id=None, load_balancer_id=None, protocol=None,
                 protocol_port=None, connection_limit=None,
                 enabled=None, provisioning_status=None, operating_status=None,
                 tls_certificate_id=None, stats=None, default_pool=None,
                 load_balancer=None, sni_containers=None, peer_port=None,
                 l7policies=None, pools=None, insert_headers=None,
                 created_at=None, updated_at=None):
        self.id = id
        self.project_id = project_id
        self.name = name
        self.description = description
        self.default_pool_id = default_pool_id
        self.load_balancer_id = load_balancer_id
        self.protocol = protocol
        self.protocol_port = protocol_port
        self.connection_limit = connection_limit
        self.enabled = enabled
        self.provisioning_status = provisioning_status
        self.operating_status = operating_status
        self.tls_certificate_id = tls_certificate_id
        self.stats = stats
        self.default_pool = default_pool
        self.load_balancer = load_balancer
        self.sni_containers = sni_containers or []
        self.peer_port = peer_port
        self.l7policies = l7policies or []
        self.insert_headers = insert_headers or {}
        self.pools = pools or []
        self.created_at = created_at
        self.updated_at = updated_at

    def update(self, update_dict):
        for key, value in update_dict.items():
            setattr(self, key, value)
            if key == 'default_pool_id':
                if self.default_pool is not None:
                    l7_pool_ids = [p.redirect_pool_id for p in self.l7policies
                                   if p.redirect_pool_id is not None and
                                   len(p.l7rules) > 0 and p.enabled is True]
                    old_pool = self.default_pool
                    if old_pool.id not in l7_pool_ids:
                        self.pools.remove(old_pool)
                        old_pool.listeners.remove(self)
                if value is not None:
                    pool = self._find_in_graph('Pool' + value)
                    if pool not in self.pools:
                        self.pools.append(pool)
                    if self not in pool.listeners:
                        pool.listeners.append(self)
                else:
                    pool = None
                setattr(self, 'default_pool', pool)

    def delete(self):
        for listener in self.load_balancer.listeners:
            if listener.id == self.id:
                self.load_balancer.listeners.remove(listener)
                break
        for pool in self.pools:
            pool.listeners.remove(self)


class LoadBalancer(BaseDataModel):

    def __init__(self, id=None, project_id=None, name=None, description=None,
                 provisioning_status=None, operating_status=None, enabled=None,
                 topology=None, vip=None, listeners=None, amphorae=None,
                 pools=None, vrrp_group=None, server_group_id=None,
                 created_at=None, updated_at=None):

        self.id = id
        self.project_id = project_id
        self.name = name
        self.description = description
        self.provisioning_status = provisioning_status
        self.operating_status = operating_status
        self.enabled = enabled
        self.vip = vip
        self.vrrp_group = vrrp_group
        self.topology = topology
        self.listeners = listeners or []
        self.amphorae = amphorae or []
        self.pools = pools or []
        self.server_group_id = server_group_id
        self.created_at = created_at
        self.updated_at = updated_at

    def update(self, update_dict):
        for key, value in update_dict.items():
            if key == 'vip':
                if self.vip is not None:
                    self.vip.update(value)
                else:
                    value.update({'load_balancer_id': self.id})
                    self.vip = Vip(**value)
            else:
                setattr(self, key, value)


class VRRPGroup(BaseDataModel):

    def __init__(self, load_balancer_id=None, vrrp_group_name=None,
                 vrrp_auth_type=None, vrrp_auth_pass=None, advert_int=None,
                 smtp_server=None, smtp_connect_timeout=None,
                 load_balancer=None):
        self.load_balancer_id = load_balancer_id
        self.vrrp_group_name = vrrp_group_name
        self.vrrp_auth_type = vrrp_auth_type
        self.vrrp_auth_pass = vrrp_auth_pass
        self.advert_int = advert_int
        self.load_balancer = load_balancer


class Vip(BaseDataModel):

    def __init__(self, load_balancer_id=None, ip_address=None,
                 subnet_id=None, network_id=None, port_id=None,
                 load_balancer=None, qos_policy_id=None):
        self.load_balancer_id = load_balancer_id
        self.ip_address = ip_address
        self.subnet_id = subnet_id
        self.network_id = network_id
        self.port_id = port_id
        self.load_balancer = load_balancer
        self.qos_policy_id = qos_policy_id


class SNI(BaseDataModel):

    def __init__(self, listener_id=None, position=None, listener=None,
                 tls_container_id=None):
        self.listener_id = listener_id
        self.position = position
        self.listener = listener
        self.tls_container_id = tls_container_id


class TLSContainer(BaseDataModel):

    def __init__(self, id=None, primary_cn=None, certificate=None,
                 private_key=None, passphrase=None, intermediates=None):
        self.id = id
        self.primary_cn = primary_cn
        self.certificate = certificate
        self.private_key = private_key
        self.passphrase = passphrase
        self.intermediates = intermediates or []


class Amphora(BaseDataModel):

    def __init__(self, id=None, load_balancer_id=None, compute_id=None,
                 status=None, lb_network_ip=None, vrrp_ip=None,
                 ha_ip=None, vrrp_port_id=None, ha_port_id=None,
                 load_balancer=None, role=None, cert_expiration=None,
                 cert_busy=False, vrrp_interface=None, vrrp_id=None,
                 vrrp_priority=None, cached_zone=None):
        self.id = id
        self.load_balancer_id = load_balancer_id
        self.compute_id = compute_id
        self.status = status
        self.lb_network_ip = lb_network_ip
        self.vrrp_ip = vrrp_ip
        self.ha_ip = ha_ip
        self.vrrp_port_id = vrrp_port_id
        self.ha_port_id = ha_port_id
        self.role = role
        self.vrrp_interface = vrrp_interface
        self.vrrp_id = vrrp_id
        self.vrrp_priority = vrrp_priority
        self.load_balancer = load_balancer
        self.cert_expiration = cert_expiration
        self.cert_busy = cert_busy
        self.cached_zone = cached_zone

    def delete(self):
        for amphora in self.load_balancer.amphorae:
            if amphora.id == self.id:
                self.load_balancer.amphorae.remove(amphora)
                break


class AmphoraHealth(BaseDataModel):

    def __init__(self, amphora_id=None, last_update=None, busy=False):
        self.amphora_id = amphora_id
        self.last_update = last_update
        self.busy = busy


class L7Rule(BaseDataModel):

    def __init__(self, id=None, l7policy_id=None, type=None, enabled=None,
                 compare_type=None, key=None, value=None, l7policy=None,
                 invert=False, provisioning_status=None, operating_status=None,
                 project_id=None, created_at=None, updated_at=None):
        self.id = id
        self.l7policy_id = l7policy_id
        self.type = type
        self.compare_type = compare_type
        self.key = key
        self.value = value
        self.l7policy = l7policy
        self.invert = invert
        self.provisioning_status = provisioning_status
        self.operating_status = operating_status
        self.project_id = project_id
        self.created_at = created_at
        self.updated_at = updated_at
        self.enabled = enabled

    def delete(self):
        if len(self.l7policy.l7rules) == 1:
            # l7policy should disappear from pool and listener lists. Since
            # we are operating only on the data model, we can fake this by
            # calling the policy's delete method.
            self.l7policy.delete()
        for r in self.l7policy.l7rules:
            if r.id == self.id:
                self.l7policy.l7rules.remove(r)
                break


class L7Policy(BaseDataModel):

    def __init__(self, id=None, name=None, description=None, listener_id=None,
                 action=None, redirect_pool_id=None, redirect_url=None,
                 position=None, listener=None, redirect_pool=None,
                 enabled=None, l7rules=None, provisioning_status=None,
                 operating_status=None, project_id=None, created_at=None,
                 updated_at=None):
        self.id = id
        self.name = name
        self.description = description
        self.listener_id = listener_id
        self.action = action
        self.redirect_pool_id = redirect_pool_id
        self.redirect_url = redirect_url
        self.position = position
        self.listener = listener
        self.redirect_pool = redirect_pool
        self.enabled = enabled
        self.l7rules = l7rules or []
        self.provisioning_status = provisioning_status
        self.operating_status = operating_status
        self.project_id = project_id
        self.created_at = created_at
        self.updated_at = updated_at

    def _conditionally_remove_pool_links(self, pool):
        """Removes links to the given pool from parent objects.

        Note this only happens if our listener isn't referencing the pool
        via its default_pool or another active l7policy's redirect_pool_id.
        """
        if (self.listener.default_pool is not None and
                pool is not None and
                pool.id != self.listener.default_pool.id and
                pool in self.listener.pools):
            listener_l7pools = [
                p.redirect_pool for p in self.listener.l7policies
                if p.redirect_pool is not None and
                len(p.l7rules) > 0 and p.enabled is True and
                p.id != self.id]
            if pool not in listener_l7pools:
                self.listener.pools.remove(pool)
                pool.listeners.remove(self.listener)

    def update(self, update_dict):
        for key, value in update_dict.items():
            if key == 'redirect_pool_id' and value is not None:
                self._conditionally_remove_pool_links(self.redirect_pool)
                self.action = constants.L7POLICY_ACTION_REDIRECT_TO_POOL
                self.redirect_url = None
                pool = self._find_in_graph('Pool' + value)
                self.redirect_pool = pool
                if len(self.l7rules) > 0 and (self.enabled is True or (
                        'enabled' in update_dict.keys() and
                        update_dict['enabled'] is True)):
                    if pool not in self.listener.pools:
                        self.listener.pools.append(pool)
                    if self.listener not in pool.listeners:
                        pool.listeners.append(self.listener)
            elif key == 'redirect_url' and value is not None:
                self.action = constants.L7POLICY_ACTION_REDIRECT_TO_URL
                self._conditionally_remove_pool_links(self.redirect_pool)
                self.redirect_pool = None
                self.redirect_pool_id = None
            elif key == 'action' and value == constants.L7POLICY_ACTION_REJECT:
                self.redirect_url = None
                self._conditionally_remove_pool_links(self.redirect_pool)
                self.redirect_pool = None
                self.redirect_pool_id = None
            elif key == 'position':
                self.listener.l7policies.remove(self)
                self.listener.l7policies.insert(value - 1, self)
            elif key == 'enabled':
                if (value is True and self.action ==
                        constants.L7POLICY_ACTION_REDIRECT_TO_POOL and
                        self.redirect_pool is not None and
                        len(self.l7rules) > 0 and
                        self.redirect_pool not in self.listener.pools):
                    self.listener.pools.append(self.redirect_pool)
                    self.redirect_pool.listeners.append(self.listener)
                elif (value is False and self.action ==
                        constants.L7POLICY_ACTION_REDIRECT_TO_POOL and
                        self.redirect_pool is not None):
                    self._conditionally_remove_pool_links(
                        self.redirect_pool)
            setattr(self, key, value)

    def delete(self):
        self._conditionally_remove_pool_links(self.redirect_pool)
        if self.redirect_pool:
            for p in self.redirect_pool.l7policies:
                if p.id == self.id:
                    self.redirect_pool.l7policies.remove(p)
        for p in self.listener.l7policies:
            if p.id == self.id:
                self.listener.l7policies.remove(p)
                break


class Quotas(BaseDataModel):

    def __init__(self,
                 project_id=None,
                 load_balancer=None,
                 listener=None,
                 pool=None,
                 health_monitor=None,
                 member=None,
                 in_use_health_monitor=None,
                 in_use_listener=None,
                 in_use_load_balancer=None,
                 in_use_member=None,
                 in_use_pool=None):
        self.project_id = project_id
        self.health_monitor = health_monitor
        self.listener = listener
        self.load_balancer = load_balancer
        self.pool = pool
        self.member = member
        self.in_use_health_monitor = in_use_health_monitor
        self.in_use_listener = in_use_listener
        self.in_use_load_balancer = in_use_load_balancer
        self.in_use_member = in_use_member
        self.in_use_pool = in_use_pool



# Copyright (c) 2014 Rackspace US, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
import stat
import uuid

from oslo_config import cfg
from oslo_log import log as logging

from octavia.certificates.common import local as local_common
from octavia.certificates.manager import cert_mgr
from octavia.common import exceptions

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class LocalCertManager(cert_mgr.CertManager):
    """Cert Manager Interface that stores data locally."""

    @staticmethod
    def store_cert(context, certificate, private_key, intermediates=None,
                   private_key_passphrase=None, **kwargs):
        """Stores (i.e., registers) a cert with the cert manager.

        This method stores the specified cert to the filesystem and returns
        a UUID that can be used to retrieve it.

        :param context: Ignored in this implementation
        :param certificate: PEM encoded TLS certificate
        :param private_key: private key for the supplied certificate
        :param intermediates: ordered and concatenated intermediate certs
        :param private_key_passphrase: optional passphrase for the supplied key

        :returns: the UUID of the stored cert
        :raises CertificateStorageException: if certificate storage fails
        """
        cert_ref = str(uuid.uuid4())
        filename_base = os.path.join(CONF.certificates.storage_path, cert_ref)

        LOG.info("Storing certificate data on the local filesystem.")
        try:
            filename_certificate = "{0}.crt".format(filename_base)
            flags = os.O_WRONLY | os.O_CREAT
            mode = stat.S_IRUSR | stat.S_IWUSR  # mode 0600
            with os.fdopen(os.open(
                    filename_certificate, flags, mode), 'w') as cert_file:
                cert_file.write(certificate)

            filename_private_key = "{0}.key".format(filename_base)
            with os.fdopen(os.open(
                    filename_private_key, flags, mode), 'w') as key_file:
                key_file.write(private_key)

            if intermediates:
                filename_intermediates = "{0}.int".format(filename_base)
                with os.fdopen(os.open(
                        filename_intermediates, flags, mode), 'w') as int_file:
                    int_file.write(intermediates)

            if private_key_passphrase:
                filename_pkp = "{0}.pass".format(filename_base)
                with os.fdopen(os.open(
                        filename_pkp, flags, mode), 'w') as pass_file:
                    pass_file.write(private_key_passphrase)
        except IOError as ioe:
            LOG.error("Failed to store certificate.")
            raise exceptions.CertificateStorageException(message=ioe.message)

        return cert_ref

    @staticmethod
    def get_cert(context, cert_ref, **kwargs):
        """Retrieves the specified cert.

        :param context: Ignored in this implementation
        :param cert_ref: the UUID of the cert to retrieve

        :return: octavia.certificates.common.Cert representation of the
                 certificate data
        :raises CertificateStorageException: if certificate retrieval fails
        """
        LOG.info("Loading certificate %s from the local filesystem.", cert_ref)

        filename_base = os.path.join(CONF.certificates.storage_path, cert_ref)

        filename_certificate = "{0}.crt".format(filename_base)
        filename_private_key = "{0}.key".format(filename_base)
        filename_intermediates = "{0}.int".format(filename_base)
        filename_pkp = "{0}.pass".format(filename_base)

        cert_data = dict()

        flags = os.O_RDONLY
        try:
            with os.fdopen(os.open(filename_certificate, flags)) as cert_file:
                cert_data['certificate'] = cert_file.read()
        except IOError:
            LOG.error("Failed to read certificate for %s.", cert_ref)
            raise exceptions.CertificateStorageException(
                msg="Certificate could not be read.")
        try:
            with os.fdopen(os.open(filename_private_key, flags)) as key_file:
                cert_data['private_key'] = key_file.read()
        except IOError:
            LOG.error("Failed to read private key for %s", cert_ref)
            raise exceptions.CertificateStorageException(
                msg="Private Key could not be read.")

        try:
            with os.fdopen(os.open(filename_intermediates, flags)) as int_file:
                cert_data['intermediates'] = int_file.read()
        except IOError:
            pass

        try:
            with os.fdopen(os.open(filename_pkp, flags)) as pass_file:
                cert_data['private_key_passphrase'] = pass_file.read()
        except IOError:
            pass

        return local_common.LocalCert(**cert_data)

    @staticmethod
    def delete_cert(context, cert_ref, **kwargs):
        """Deletes the specified cert.

        :param context: Ignored in this implementation
        :param cert_ref: the UUID of the cert to delete

        :raises CertificateStorageException: if certificate deletion fails
        """
        LOG.info("Deleting certificate %s from the local filesystem.",
                 cert_ref)

        filename_base = os.path.join(CONF.certificates.storage_path, cert_ref)

        filename_certificate = "{0}.crt".format(filename_base)
        filename_private_key = "{0}.key".format(filename_base)
        filename_intermediates = "{0}.int".format(filename_base)
        filename_pkp = "{0}.pass".format(filename_base)

        try:
            os.remove(filename_certificate)
            os.remove(filename_private_key)
            os.remove(filename_intermediates)
            os.remove(filename_pkp)
        except IOError as ioe:
            LOG.error("Failed to delete certificate %s", cert_ref)
            raise exceptions.CertificateStorageException(message=ioe.message)



# Copyright (c) 2014 Rackspace US, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Common classes for Barbican certificate handling
"""

import abc

from barbicanclient.v1 import containers
from oslo_utils import encodeutils
import six

from octavia.certificates.common import cert
from octavia.common.tls_utils import cert_parser
from octavia.i18n import _


class BarbicanCert(cert.Cert):
    """Representation of a Cert based on the Barbican CertificateContainer."""
    def __init__(self, cert_container):
        if not isinstance(cert_container, containers.CertificateContainer):
            raise TypeError(_("Retrieved Barbican Container is not of the "
                              "correct type (certificate)."))
        self._cert_container = cert_container

    def get_certificate(self):
        if self._cert_container.certificate:
            return encodeutils.to_utf8(
                self._cert_container.certificate.payload)

    def get_intermediates(self):
        if self._cert_container.intermediates:
            intermediates = encodeutils.to_utf8(
                self._cert_container.intermediates.payload)
            return [imd for imd in cert_parser.get_intermediates_pems(
                intermediates)]

    def get_private_key(self):
        if self._cert_container.private_key:
            return encodeutils.to_utf8(
                self._cert_container.private_key.payload)

    def get_private_key_passphrase(self):
        if self._cert_container.private_key_passphrase:
            return encodeutils.to_utf8(
                self._cert_container.private_key_passphrase.payload)


@six.add_metaclass(abc.ABCMeta)
class BarbicanAuth(object):
    @abc.abstractmethod
    def get_barbican_client(self, project_id):
        """Creates a Barbican client object.

        :param project_id: Project ID that the request will be used for
        :return: a Barbican Client object
        :raises Exception: if the client cannot be created
        """



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""update member address column

Revision ID: 4faaa983e7a9
Revises: 13500e2e978d
Create Date: 2014-09-29 11:22:16.565071

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4faaa983e7a9'
down_revision = '13500e2e978d'


def upgrade():
    op.alter_column(u'member', u'address', new_column_name=u'ip_address',
                    existing_type=sa.String(64))



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import octavia.common.data_models as models


class Topology(models.BaseDataModel):

    def __init__(self, hostname=None, uuid=None, topology=None, role=None,
                 ip=None, ha_ip=None):
        self.hostname = hostname
        self.uuid = uuid
        self.topology = topology
        self.role = role
        self.ip = ip
        self.ha_ip = ha_ip


class Info(models.BaseDataModel):

    def __init__(self, hostname=None, uuid=None, version=None,
                 api_version=None):
        self.hostname = hostname
        self.uuid = uuid
        self.version = version
        self.api_version = api_version


class Details(models.BaseDataModel):

    def __init__(self, hostname=None, uuid=None, version=None,
                 api_version=None, network_tx=None, network_rx=None,
                 active=None, haproxy_count=None, cpu=None, memory=None,
                 disk=None, load=None, listeners=None, packages=None):
        self.hostname = hostname
        self.uuid = uuid,
        self.version = version
        self.api_version = api_version
        self.network_tx = network_tx
        self.network_rx = network_rx
        self.active = active
        self.haproxy_count = haproxy_count
        self.cpu = cpu
        self.memory = memory
        self.disk = disk
        self.load = load or []
        self.listeners = listeners or []
        self.packages = packages or []


class CPU(models.BaseDataModel):

    def __init__(self, total=None, user=None, system=None, soft_irq=None):
        self.total = total
        self.user = user
        self.system = system
        self.soft_irq = soft_irq


class Memory(models.BaseDataModel):

    def __init__(self, total=None, free=None, available=None, buffers=None,
                 cached=None, swap_used=None, shared=None, slab=None,
                 committed_as=None):
        self.total = total
        self.free = free
        self.available = available
        self.buffers = buffers
        self.cached = cached
        self.swap_used = swap_used
        self.shared = shared
        self.slab = slab
        self.committed_as = committed_as


class Disk(models.BaseDataModel):

    def __init__(self, used=None, available=None):
        self.used = used
        self.available = available


class ListenerStatus(models.BaseDataModel):

    def __init__(self, status=None, uuid=None, provisioning_status=None,
                 type=None, pools=None):
        self.status = status
        self.uuid = uuid
        self.provisioning_status = provisioning_status
        self.type = type
        self.pools = pools or []


class Pool(models.BaseDataModel):

    def __init__(self, uuid=None, status=None, members=None):
        self.uuid = uuid
        self.status = status
        self.members = members or []



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia.api.common import hooks

# Pecan Application Configurations
# See https://pecan.readthedocs.org/en/latest/configuration.html#application-configuration # noqa
app = {
    'root': 'octavia.api.root_controller.RootController',
    'modules': ['octavia.api'],
    'hooks': [
        hooks.ContextHook(),
        hooks.QueryParametersHook()],
    'debug': False
}

# WSME Configurations
# See https://wsme.readthedocs.org/en/latest/integrate.html#configuration
wsme = {
    # Keeping debug True for now so we can easily troubleshoot.
    'debug': True
}



# Copyright 2016 Rackspace US Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from six.moves.urllib import parse
from tempest.lib.common import rest_client


class MembersClient(rest_client.RestClient):
    """Tests Members API."""

    _MEMBERS_URL = ("v1/loadbalancers/{lb_id}/pools/{pool_id}/members")
    _MEMBER_URL = "{base_url}/{{member_id}}".format(base_url=_MEMBERS_URL)

    def list_members(self, lb_id, pool_id, params=None):
        """List all Members."""
        url = self._MEMBERS_URL.format(lb_id=lb_id, pool_id=pool_id)
        if params:
            url = "{0}?{1}".format(url, parse.urlencode(params))
        resp, body = self.get(url)
        body = jsonutils.loads(body)
        self.expected_success(200, resp.status)
        return rest_client.ResponseBodyList(resp, body)

    def get_member(self, lb_id, pool_id, member_id, params=None):
        """Get member details."""
        url = self._MEMBER_URL.format(lb_id=lb_id,
                                      pool_id=pool_id,
                                      member_id=member_id)
        if params:
            url = '{0}?{1}'.format(url, parse.urlencode(params))
        resp, body = self.get(url)
        body = jsonutils.loads(body)
        self.expected_success(200, resp.status)
        return rest_client.ResponseBody(resp, body)

    def create_member(self, lb_id, pool_id, **kwargs):
        """Create member."""
        url = self._MEMBERS_URL.format(lb_id=lb_id,
                                       pool_id=pool_id)
        post_body = jsonutils.dumps(kwargs)
        resp, body = self.post(url, post_body)
        body = jsonutils.loads(body)
        self.expected_success(202, resp.status)
        return rest_client.ResponseBody(resp, body)

    def update_member(self, lb_id, pool_id, member_id, **kwargs):
        """Update member."""
        url = self._MEMBER_URL.format(lb_id=lb_id,
                                      pool_id=pool_id,
                                      member_id=member_id)
        put_body = jsonutils.dumps(kwargs)
        resp, body = self.put(url, put_body)
        body = jsonutils.loads(body)
        self.expected_success(202, resp.status)
        return rest_client.ResponseBody(resp, body)

    def delete_member(self, lb_id, pool_id, member_id):
        """Delete member."""
        url = self._MEMBER_URL.format(lb_id=lb_id,
                                      pool_id=pool_id,
                                      member_id=member_id)
        resp, body = self.delete(url)
        self.expected_success(202, resp.status)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils

from octavia.common import constants
from octavia.common import data_models
from octavia.compute import compute_base as driver_base

LOG = logging.getLogger(__name__)


class NoopManager(object):
    def __init__(self):
        super(NoopManager, self).__init__()
        self.computeconfig = {}

    def build(self, name="amphora_name", amphora_flavor=None,
              image_id=None, image_tag=None, image_owner=None,
              key_name=None, sec_groups=None, network_ids=None,
              config_drive_files=None, user_data=None, port_ids=None,
              server_group_id=None):
        LOG.debug("Compute %s no-op, build name %s, amphora_flavor %s, "
                  "image_id %s, image_tag %s, image_owner %s, key_name %s, "
                  "sec_groups %s, network_ids %s, config_drive_files %s, "
                  "user_data %s, port_ids %s, server_group_id %s",
                  self.__class__.__name__,
                  name, amphora_flavor, image_id, image_tag, image_owner,
                  key_name, sec_groups, network_ids, config_drive_files,
                  user_data, port_ids, server_group_id)
        self.computeconfig[(name, amphora_flavor, image_id, image_tag,
                            image_owner, key_name, user_data,
                            server_group_id)] = (
            name, amphora_flavor,
            image_id, image_tag, image_owner, key_name, sec_groups,
            network_ids, config_drive_files, user_data, port_ids,
            server_group_id, 'build')
        compute_id = uuidutils.generate_uuid()
        return compute_id

    def delete(self, compute_id):
        LOG.debug("Compute %s no-op, compute_id %s",
                  self.__class__.__name__, compute_id)
        self.computeconfig[compute_id] = (compute_id, 'delete')

    def status(self, compute_id):
        LOG.debug("Compute %s no-op, compute_id %s",
                  self.__class__.__name__, compute_id)
        self.computeconfig[compute_id] = (compute_id, 'status')
        return constants.UP

    def get_amphora(self, compute_id):
        LOG.debug("Compute %s no-op, compute_id %s",
                  self.__class__.__name__, compute_id)
        self.computeconfig[compute_id] = (compute_id, 'get_amphora')
        return data_models.Amphora(
            compute_id=compute_id,
            status=constants.ACTIVE,
            lb_network_ip='192.0.2.1'
        ), None

    def create_server_group(self, name, policy):
        LOG.debug("Create Server Group %s no-op, name %s, policy %s ",
                  self.__class__.__name__, name, policy)
        self.computeconfig[(name, policy)] = (name, policy, 'create')

    def delete_server_group(self, server_group_id):
        LOG.debug("Delete Server Group %s no-op, id %s ",
                  self.__class__.__name__, server_group_id)
        self.computeconfig[server_group_id] = (server_group_id, 'delete')


class NoopComputeDriver(driver_base.ComputeBase):
    def __init__(self):
        super(NoopComputeDriver, self).__init__()
        self.driver = NoopManager()

    def build(self, name="amphora_name", amphora_flavor=None,
              image_id=None, image_tag=None, image_owner=None,
              key_name=None, sec_groups=None, network_ids=None,
              config_drive_files=None, user_data=None, port_ids=None,
              server_group_id=None):

        compute_id = self.driver.build(name, amphora_flavor,
                                       image_id, image_tag, image_owner,
                                       key_name, sec_groups, network_ids,
                                       config_drive_files, user_data, port_ids,
                                       server_group_id)
        return compute_id

    def delete(self, compute_id):
        self.driver.delete(compute_id)

    def status(self, compute_id):
        return self.driver.status(compute_id)

    def get_amphora(self, compute_id):
        return self.driver.get_amphora(compute_id)

    def create_server_group(self, name, policy):
        return self.driver.create_server_group(name, policy)

    def delete_server_group(self, server_group_id):
        self.driver.delete_server_group(server_group_id)



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

import octavia.common.config as config
import octavia.tests.unit.base as base


class TestConfig(base.TestCase):

    def test_sanity(self):
        config.init([])
        config.setup_logging(cfg.CONF)
        # Resetting because this will cause inconsistent errors when run with
        # other tests
        self.addCleanup(cfg.CONF.reset)

    def test_validate_server_certs_key_passphrase(self):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))
        conf.config(
            group="certificates",
            server_certs_key_passphrase="insecure-key-do-not-use-this-key"
        )

        # Test too short
        self.assertRaises(ValueError, conf.config,
                          group="certificates",
                          server_certs_key_passphrase="short_passphrase")

        # Test too long
        self.assertRaises(
            ValueError, conf.config, group="certificates",
            server_certs_key_passphrase="long-insecure-key-do-not-use-this")

        # Test invalid characters
        self.assertRaises(
            ValueError, conf.config, group="certificates",
            server_certs_key_passphrase="insecure-key-do-not-u$e-this-key")
